import unittest

from openformats.tests.formats.common import CommonFormatTestMixin
from openformats.tests.utils import strip_leading_spaces
from openformats.formats.srt import SrtHandler


class SrtTestCase(CommonFormatTestMixin, unittest.TestCase):
    HANDLER_CLASS = SrtHandler
    TESTFILE_BASE = "openformats/tests/formats/srt/files"

    def test_srt_occurrences(self):
        """srt: Test that timings are saved as occurrencies."""
        source = strip_leading_spaces("""
            1
            00:01:28,797 --> 00:01:30,297
            Hello, World!
        """)
        template, stringset = self.handler.parse(source)
        self.assertEqual(stringset[0].occurrences, '00:01:28.797,00:01:30.297')

    def test_missing_order(self):
        source = strip_leading_spaces("""
            00:01:28,797 --> 00:01:30,297
            Hello, World!
        """)
        self._test_parse_error(
            source,
            "Not enough data on subtitle section on line 2. Order number, "
            "timings and subtitle content are needed"
        )

    def test_missing_timings(self):
        source = strip_leading_spaces("""
            1
            Hello, World!
        """)
        self._test_parse_error(
            source,
            "Not enough data on subtitle section on line 2. Order number, "
            "timings and subtitle content are needed"
        )

    def test_missing_string(self):
        source = strip_leading_spaces("""
            1
            00:01:28,797 --> 00:01:30,297
        """)
        self._test_parse_error(
            source,
            "Not enough data on subtitle section on line 2. Order number, "
            "timings and subtitle content are needed"
        )

    def test_negative_order(self):
        source = strip_leading_spaces("""
            -3
            00:01:28,797 --> 00:01:30,297
            Hello, World!
        """)
        self._test_parse_error(
            source,
            "Order number on line 2 (-3) must be a positive integer"
        )

    def test_non_ascending_order(self):
        source = strip_leading_spaces("""
            2
            00:01:28,797 --> 00:01:30,297
            Hello, World!

            1
            00:04:28,797 --> 00:08:30,297
            Goodbye, World!
        """)
        self._test_parse_error(
            source,
            "Order numbers must be in ascending order; number in line 6 (1) "
            "is wrong"
        )

    def test_wrong_timings(self):
        source = strip_leading_spaces("""
            1
            00:01:28,797 00:01:30,297
            Hello, World!
        """)
        self._test_parse_error(
            source,
            "Timings on line 3 don't follow '[start] --> [end] (position)' "
            "pattern"
        )

        source = strip_leading_spaces("""
            1
            00:fas28,797 --> 00:01:30,297
            Hello, World!
        """)
        self._test_parse_error(
            source,
            "Problem with start of timing at line 3: '00:fas28,797'"
        )

        source = strip_leading_spaces("""
            1
            00:01:28,797 --> 00:ois30,297
            Hello, World!
        """)
        self._test_parse_error(
            source,
            "Problem with end of timing at line 3: '00:ois30,297'"
        )



import json
from hashlib import md5

from django.db import models
from django.core.urlresolvers import reverse


class Payload(models.Model):
    payload_hash = models.CharField(max_length=32, unique=True,
                                    blank=True, null=True)
    payload = models.TextField()
    created = models.DateTimeField(auto_now_add=True)
    last_viewed = models.DateTimeField(blank=True, null=True)

    def _presave(self):
        self.payload_hash = md5(self.payload.encode('utf-8')).hexdigest()

    def save(self):
        self._presave()
        super(Payload, self).save()

    def get_absolute_url(self):
        return reverse("testbed_main",
                       kwargs={'payload_hash': self.payload_hash})

    def __unicode__(self):
        payload = json.loads(self.payload)
        handler = payload.get('handler', None)
        if handler:
            return u"{}-{}".format(handler, self.payload_hash[-8:])
        else:
            return self.payload_hash[-8:]



import re

from openformats.transcribers import Transcriber


class OrderedCompilerMixin(object):
    SPACE_PAT = re.compile(r'^\s*$')

    def compile(self, template, stringset):
        # assume stringset is ordered within the template
        transcriber = Transcriber(template)
        template = transcriber.source

        for string in stringset:
            hash_position = template.index(string.template_replacement)
            if not string.pluralized:
                transcriber.copy_until(hash_position)
                transcriber.add(string.string)
                transcriber.skip(len(string.template_replacement))
            else:
                # if the hash is on its own on a line with only spaces, we have
                # to remember it's indent
                indent_length = template[hash_position::-1].index('\n') - 1
                indent = template[hash_position - indent_length:hash_position]
                tail_length = template[
                    hash_position + len(string.template_replacement):
                ].index('\n')
                tail = template[
                    hash_position + len(string.template_replacement):
                    hash_position + len(string.template_replacement) +
                    tail_length
                ]
                if (self.SPACE_PAT.search(indent) and
                        self.SPACE_PAT.search(tail)):
                    transcriber.copy_until(hash_position - indent_length)
                    for rule, value in string.string.items():
                        transcriber.add(
                            indent + self.plural_template.format(
                                rule=self.RULES_ITOA[rule], string=value
                            ) + tail + '\n'
                        )
                    transcriber.skip(indent_length +
                                     len(string.template_replacement) +
                                     tail_length + 1)
                else:
                    # string is not on its own, simply replace hash with all
                    # plural forms
                    transcriber.copy_until(hash_position)
                    for rule, value in string.string.items():
                        transcriber.add(self.plural_template.format(
                            rule=self.RULES_ITOA[rule], string=value
                        ))
                    transcriber.skip(len(string.template_replacement))

        transcriber.copy_until(len(template))
        compiled = transcriber.get_destination()

        return compiled



from __future__ import absolute_import

import re

from ..handlers import Handler
from openformats.strings import OpenString
from openformats.transcribers import Transcriber
from ..utils.xml import DumbXml


class BetaAndroidHandler(Handler):
    name = "BETA_ANDROID"
    extension = "xml"

    plural_template = u'<item quantity="{rule}">{string}</item>'
    SPACE_PAT = re.compile(r'^\s*$')
    # Atttibutes that designate a string should be filtered out
    FILTER_ATTRIBUTES = {
        'translatable': 'false'
    }

    EXTRACTS_RAW = False

    def parse(self, content):
        stringset = []
        if type(content) == str:
            content = content.decode("utf-8")  # convert to unicode

        resources_tag_position = content.index("<resources")

        self.transcriber = Transcriber(content[resources_tag_position:])
        source = self.transcriber.source

        self._order = 0

        resources_tag = DumbXml(source)
        last_comment = ""
        for tag, offset in resources_tag.find(("string-array", "string",
                                               "plurals", DumbXml.COMMENT)):
            if self._should_ignore(tag):
                last_comment = ""
                continue
            if tag.name == DumbXml.COMMENT:
                last_comment = tag.inner
                self.transcriber.copy_until(offset + len(tag.content))
            elif tag.name == "string":
                string = self._handle_string_tag(tag, offset, last_comment)
                last_comment = ""
                if string is not None:
                    stringset.append(string)
            elif tag.name == "string-array":
                for string in self._handle_string_array_tag(tag, offset,
                                                            last_comment):
                    if string is not None:
                        stringset.append(string)
                last_comment = ""
            elif tag.name == "plurals":
                string = self._handle_plurals_tag(tag, offset, last_comment)
                if string is not None:
                    stringset.append(string)
                last_comment = ""

        self.transcriber.copy_until(len(source))

        template = content[:resources_tag_position] +\
            self.transcriber.get_destination()

        self.transcriber = None

        return template, stringset

    def _handle_string_tag(self, tag, offset, comment):
        string = None
        if tag.inner.strip() != "":
            context = tag.attrs.get('product', "")
            string = OpenString(tag.attrs['name'], tag.inner,
                                context=context, order=self._order,
                                developer_comment=comment)
            self._order += 1

        # ... <string name="foo">Hello ....
        #                        ^
        self.transcriber.copy_until(offset + tag.inner_offset)

        # ... ing name="foo">Hello world</stri...
        #                               ^
        if string is not None:
            self.transcriber.add(string.template_replacement)
            self.transcriber.skip(len(tag.inner))
        else:
            self.transcriber.copy_until(offset + tag.inner_offset +
                                        len(tag.inner))

        # ...ello World</string>
        #                       ^
        self.transcriber.copy_until(offset + len(tag.content))

        return string

    def _handle_string_array_tag(self, string_array_tag, string_array_offset,
                                 comment):
        # ...ing-array>   <item>H...
        #              ^
        self.transcriber.copy_until(string_array_offset +
                                    string_array_tag.inner_offset)

        context = string_array_tag.attrs.get('product', "")
        for index, (item_tag, item_offset) in enumerate(
                string_array_tag.find('item')):
            string = None
            if item_tag.inner.strip() != "":
                string = OpenString(
                    "{}[{}]".format(string_array_tag.attrs['name'], index),
                    item_tag.inner,
                    context=context,
                    order=self._order,
                    developer_comment=comment
                )
                self._order += 1
                yield string

            # ... <item>Hello...
            #           ^
            self.transcriber.copy_until(string_array_offset + item_offset +
                                        item_tag.inner_offset)

            # ...ello world</item>...
            #              ^
            if string is not None:
                self.transcriber.add(string.template_replacement)
                self.transcriber.skip(len(item_tag.inner))
            else:
                self.transcriber.copy_until(string_array_offset + item_offset +
                                            item_tag.inner_offset)

            # orld</item>   <it...
            #            ^
            self.transcriber.copy_until(
                string_array_offset + item_offset + item_tag.inner_offset +
                len(item_tag.content)
            )

        # </item>  </string-array>
        #                         ^
        self.transcriber.copy_until(string_array_offset +
                                    len(string_array_tag.content))

    def _handle_plurals_tag(self, plurals_tag, plurals_offset, comment):
        # <plurals name="foo">   <item>Hello ...
        #                     ^
        self.transcriber.copy_until(plurals_offset + plurals_tag.inner_offset)

        first_item_offset = None
        strings = {}
        for item_tag, item_offset in plurals_tag.find('item'):
            if item_tag.inner.strip() == "":
                strings = None
                break

            first_item_offset = first_item_offset or item_offset

            rule = self.get_rule_number(item_tag.attrs['quantity'])
            strings[rule] = item_tag.inner
        last_item_tag, last_item_offset = item_tag, item_offset

        if strings is not None:
            context = plurals_tag.attrs.get('product', "")
            string = OpenString(plurals_tag.attrs['name'], strings,
                                pluralized=True,
                                context=context, order=self._order,
                                developer_comment=comment)
            self._order += 1

            # <plurals name="foo">   <item>Hello ...
            #                        ^
            self.transcriber.copy_until(plurals_offset + first_item_offset)

            # ...</item>   </plurals>...
            #           ^
            self.transcriber.add(string.template_replacement)
            self.transcriber.skip(last_item_offset +
                                  len(last_item_tag.content) -
                                  first_item_offset)

        else:
            string = None

        # ...</plurals> ...
        #              ^
        self.transcriber.copy_until(plurals_offset + len(plurals_tag.content))

        return string

    def _should_ignore(self, tag):
        """
        If the tag has a key: value elemement that matches FILTER_ATTRIBUTES
        it will return True, else it returns False
        """
        for key, value in self.FILTER_ATTRIBUTES.iteritems():
            filter_attr = tag.attrs.get(key, None)
            if filter_attr is not None and filter_attr == value:
                return True
        return False

    def compile(self, template, stringset):
        resources_tag_position = template.index("<resources")
        self._stringset = list(stringset)
        self._stringset_index = 0

        self.transcriber = Transcriber(template[resources_tag_position:])
        self.source = self.transcriber.source

        resources_tag = DumbXml(self.source)

        for tag, offset in resources_tag.find(("string", "string-array",
                                               "plurals")):
            if self._should_ignore(tag):
                continue
            if tag.name == "string":
                self._compile_string(tag, offset)
            elif tag.name == "string-array":
                self._compile_string_array(tag, offset)
            elif tag.name == "plurals":
                self._compile_plurals(tag, offset)
        self.transcriber.copy_until(len(self.source))

        # Lets do another pass to clear empty <string-array>s
        self.transcriber = Transcriber(self.transcriber.get_destination())
        self.source = self.transcriber.source
        resources_tag = DumbXml(self.source)
        for string_array_tag, string_array_offset in resources_tag.find(
                "string-array"):
            if (string_array_tag.inner and
                    len(list(string_array_tag.find("item"))) == 0):
                self.transcriber.copy_until(string_array_offset)
                self.transcriber.skip(len(string_array_tag.content))
        self.transcriber.copy_until(len(self.source))

        compiled = template[:resources_tag_position] +\
            self.transcriber.get_destination()

        self._stringset = None
        self._stringset_index = None
        self.transcriber = None

        return compiled

    def _compile_string(self, string_tag, string_offset):
        try:
            next_string = self._stringset[self._stringset_index]
        except IndexError:
            next_string = None
        if (next_string is not None and
                next_string.template_replacement == string_tag.inner):
            # found one to replace
            self._stringset_index += 1

            self.transcriber.copy_until(string_offset +
                                        string_tag.inner_offset)
            self.transcriber.add(next_string.string)
            self.transcriber.skip(len(string_tag.inner))
            self.transcriber.copy_until(string_offset +
                                        len(string_tag.content))

        else:
            # didn't find it, must remove by skipping it
            self.transcriber.copy_until(string_offset)
            self.transcriber.skip(len(string_tag.content))

    def _compile_string_array(self, string_array_tag, string_array_offset):
        self.transcriber.copy_until(string_array_offset +
                                    string_array_tag.inner_offset)
        for item_tag, item_offset in string_array_tag.find("item"):
            try:
                next_string = self._stringset[self._stringset_index]
            except IndexError:
                next_string = None
            if (next_string is not None and
                    next_string.template_replacement == item_tag.inner):
                # found one to replace
                self._stringset_index += 1

                self.transcriber.copy_until(string_array_offset + item_offset +
                                            item_tag.inner_offset)
                self.transcriber.add(next_string.string)
                self.transcriber.skip(len(item_tag.inner))
                self.transcriber.copy_until(string_array_offset + item_offset +
                                            len(item_tag.content))

            else:
                # didn't find it, must remove by skipping it
                self.transcriber.copy_until(string_array_offset + item_offset)
                self.transcriber.skip(len(item_tag.content))
        self.transcriber.copy_until(string_array_offset +
                                    len(string_array_tag.content))

    def _compile_plurals(self, plurals_tag, plurals_offset):
        try:
            next_string = self._stringset[self._stringset_index]
        except IndexError:
            next_string = None
        if (next_string is not None and
                next_string.template_replacement == plurals_tag.inner.strip()):
            # found one to replace, if the hash is on its own on a line with
            # only spaces, we have to remember it's indent
            self._stringset_index += 1

            is_multiline = True
            indent_length = tail_length = 0
            try:
                hash_position = plurals_offset + plurals_tag.inner_offset +\
                    plurals_tag.inner.index(next_string.template_replacement)
                indent_length = self.source[hash_position::-1].\
                    index('\n') - 1
                indent = self.source[hash_position -
                                     indent_length:hash_position]
                end_of_hash = (hash_position +
                               len(next_string.template_replacement))
                tail_length = self.source[end_of_hash:].index('\n')
                tail = self.source[end_of_hash:end_of_hash + tail_length]
            except ValueError:
                is_multiline = False

            is_multiline = (is_multiline and
                            (self.SPACE_PAT.search(indent) and
                             self.SPACE_PAT.search(tail)))

            if is_multiline:
                # write until beginning of hash
                self.transcriber.copy_until(hash_position - indent_length)
                for rule, value in next_string.string.items():
                    self.transcriber.add(
                        indent +
                        self.plural_template.format(
                            rule=self.get_rule_string(rule), string=value
                        ) +
                        tail + '\n'
                    )
                self.transcriber.skip(indent_length +
                                      len(next_string.template_replacement) +
                                      tail_length + 1)

            else:
                # string is not on its own, simply replace hash with all plural
                # forms
                self.transcriber.copy_until(hash_position)
                for rule, value in next_string.string.items():
                    self.transcriber.add(
                        self.plural_template.format(
                            rule=self.get_rule_string(rule), string=value
                        )
                    )
                self.transcriber.skip(indent_length +
                                      len(next_string.template_replacement) +
                                      tail_length)

            # finish up by copying until the end of </plurals>
            self.transcriber.copy_until(plurals_offset +
                                        len(plurals_tag.content))

        else:
            # didn't find it, must remove by skipping it
            self.transcriber.copy_until(plurals_offset)
            self.transcriber.skip_until(plurals_offset +
                                        len(plurals_tag.content))



import unittest

from openformats.tests.formats.common import CommonFormatTestMixin
from openformats.formats.plaintext import PlaintextHandler


class PlaintextTestCase(CommonFormatTestMixin, unittest.TestCase):
    HANDLER_CLASS = PlaintextHandler
    TESTFILE_BASE = "openformats/tests/formats/plaintext/files"



from os.path import abspath, dirname
import sys

import nose


def run_all(args=None):
    if not args:
        args = [
            'nosetests', '--with-xunit', '--with-xcoverage',
            '--cover-package=openformats', '--cover-erase',
            '--logging-filter=openformats', '--logging-level=DEBUG',
            '--verbose',
        ]

    nose.run_exit(
        argv=args,
        defaultTest=abspath(dirname(__file__))
    )


if __name__ == '__main__':
    run_all(sys.argv)



import fnmatch
import re

from os import listdir, path
from os.path import isfile, join

from openformats.exceptions import ParseError
from openformats.tests.utils import translate_stringset


class CommonFormatTestMixin(object):
    """
    Define a set of tests to be run by every file format.

    The class that inherits from this must define the following:

    * ``HANDLER_CLASS``, eg: PlaintextHandler
    * ``TESTFILE_BASE``, eg: `openformats/tests/formats/plaintext/files`
    """

    TESTFILE_BASE = None
    HANDLER_CLASS = None

    def __init__(self, *args, **kwargs):
        self.data = {}
        super(CommonFormatTestMixin, self).__init__(*args, **kwargs)

    def read_files(self, ftypes=('en', 'el', 'tpl')):
        """
        Read test data files into variables.

        Example: 1_en.txt stored into self.data["1_en"]
        """

        # Find source files to use as a base to read all others
        en_files = []
        for f in listdir(self.TESTFILE_BASE):
            if (isfile(join(self.TESTFILE_BASE, f)) and
                    fnmatch.fnmatch(f, '[!.]*_en.*')):
                en_files.append(f)

        file_nums = set([f.split("_")[0] for f in en_files])
        for num in file_nums:
            for ftype in ftypes:
                name = "%s_%s" % (num, ftype)  # 1_en, 1_fr etc.
                filepath = path.join(self.TESTFILE_BASE, "%s.%s" % (
                    name, self.HANDLER_CLASS.extension))
                if not isfile(filepath):
                    self.fail("Bad test files: Expected to find %s" % filepath)
                with open(filepath, "r") as myfile:
                    self.data[name] = myfile.read().decode("utf-8")

    def setUp(self):
        self.handler = self.HANDLER_CLASS()
        self.read_files()
        self.tmpl, self.strset = self.handler.parse(self.data["1_en"])
        super(CommonFormatTestMixin, self).setUp()

    def test_template(self):
        """Test that the template created is the same as static one."""
        # FIXME: Test descriptions should have the handler's name prefixed to
        # be able to differentiate between them.
        self.assertEquals(self.tmpl, self.data["1_tpl"])

    def test_no_empty_strings_in_handler_stringset(self):
        for s in self.strset:
            self.assertFalse(s.string == '')

    def test_compile(self):
        """Test that import-export is the same as the original file."""
        remade_orig_content = self.handler.compile(self.tmpl, self.strset)
        self.assertEquals(remade_orig_content, self.data["1_en"])

    def test_translate(self):
        """Test that translate + export is the same as the precompiled file."""
        translated_strset = translate_stringset(self.strset)
        translated_content = self.handler.compile(self.tmpl, translated_strset)
        self.assertEquals(translated_content, self.data["1_el"])

    def _test_parse_error(self, source, error_msg):
        """
        Test that trying to parse 'source' raises an error with a message
        exactly like 'error_msg'
        """
        self.assertRaisesRegexp(ParseError,
                                r'^{}$'.format(re.escape(error_msg)),
                                lambda: self.handler.parse(source))



import re
import unittest

from openformats.utils.xml import NewDumbXml, DumbXmlSyntaxError


class DumbXmlTestCase(unittest.TestCase):
    def test_no_embeds(self):
        cases = (
            # Without content
            ('<a/>', 0, 'a', {}, None, None, None, None, 4, '', 4),
            ('<a:b/>', 0, 'a:b', {}, None, None, None, None, 6, '', 6),
            (' <a/>', 1, 'a', {}, None, None, None, None, 5, '', 5),
            ('<a />', 0, 'a', {}, None, None, None, None, 5, '', 5),
            ('<a/ >', 0, 'a', {}, None, None, None, None, 5, '', 5),
            ('<a/> tail', 0, 'a', {}, None, None, None, None, 4, ' tail', 9),
            ('<a b="c" d:e="f"/>', 0, 'a', {'b': "c", 'd:e': "f"}, None, None,
             None, None, 18, '', 18),
            ('<a b="c" d="e" />', 0, 'a', {'b': "c", 'd': "e"}, None, None,
             None, None, 17, '', 17),
            ('<a b="c/d" e="f"/>', 0, 'a', {'b': "c/d", 'e': "f"}, None, None,
             None, None, 18, '', 18),
            ('<a b="c>d" e="f"/>', 0, 'a', {'b': "c>d", 'e': "f"}, None, None,
             None, None, 18, '', 18),

            # Comments
            ('<!---->', 0, NewDumbXml.COMMENT, {}, 4, '', '', 4, 7, '', 7),
            ('<!-- hello there -->', 0, NewDumbXml.COMMENT, {}, 4,
             ' hello there ', ' hello there ', 17, 20, '', 20),
            ('<!-- hello there --> tail', 0, NewDumbXml.COMMENT, {}, 4,
             ' hello there ', ' hello there ', 17, 20, ' tail', 25),

            # With content
            ('<a></a>', 0, 'a', {}, 3, '', '', 3, 7, '', 7),
            ('<a></a> tail', 0, 'a', {}, 3, '', '', 3, 7, ' tail', 12),
            ('<a b="c"></a> tail', 0, 'a', {'b': "c"}, 9, '', '', 9, 13,
             ' tail', 18),
            ('<a>hello world</a>', 0, 'a', {}, 3, 'hello world', 'hello world',
             14, 18, '', 18),
            ('<a>hello <![CDATA[</a>]]></a>', 0, 'a', {}, 3,
             'hello <![CDATA[</a>]]>', 'hello <![CDATA[</a>]]>', 25, 29, '',
             29),
            ('<a>hello world</a> tail', 0, 'a', {}, 3, 'hello world',
             'hello world', 14, 18, ' tail', 23),
            ('<a b="c">hello world</a> tail', 0, 'a', {'b': "c"}, 9,
             'hello world', 'hello world', 20, 24, ' tail', 29),
        )
        for case in cases:
            dumb_xml = NewDumbXml(case[0])
            self.assertEquals(
                (dumb_xml.source, dumb_xml.position, dumb_xml.tag,
                 dumb_xml.attrib, dumb_xml.text_position, dumb_xml.text,
                 dumb_xml.content, dumb_xml.content_end,
                 dumb_xml.tail_position, dumb_xml.tail, dumb_xml.end),
                case
            )

    def test_no_embeds_errors(self):
        cases = (('<a', "Opening tag not closed on line 1"),
                 ('<a b="c"', "Opening tag 'a' not closed on line 1"),
                 ('<a', "Opening tag not closed on line 1"),
                 ('<a/', "Opening tag 'a' not closed on line 1"),
                 ('<a /', "Opening tag 'a' not closed on line 1"),
                 ('<a/oiajod', "Opening tag 'a' not closed on line 1"),
                 ('<a>', "Tag 'a' not closed on line 1"),
                 ('<a>hello world', "Tag 'a' not closed on line 1"),
                 ('<!---', "Comment not closed on line 1"),
                 ('<!--jafosijdfoas-', "Comment not closed on line 1"))
        for source, error_msg in cases:
            with self.assertRaisesRegexp(DumbXmlSyntaxError,
                                         r'^{}$'.format(re.escape(error_msg))):
                dumb_xml = NewDumbXml(source)
                dumb_xml.end  # should expand all properties eventually

    def test_one_child(self):
        cases = (
            # Without content
            ('<r><a/></r>', '<a/>', 7, 3, 'a', {}, None, None, 7, '', 7),
            ('<r><a:b/></r>', '<a:b/>', 9, 3, 'a:b', {}, None, None, 9, '', 9),
            ('<r> <a/></r>', ' <a/>', 8, 4, 'a', {}, None, None, 8, '', 8),
            ('<r><a /></r>', '<a />', 8, 3, 'a', {}, None, None, 8, '', 8),
            ('<r><a/ ></r>', '<a/ >', 8, 3, 'a', {}, None, None, 8, '', 8),
            ('<r><a/> tail</r>', '<a/> tail', 12, 3, 'a', {}, None, None, 7,
             ' tail', 12),
            ('<r><a b="c" d:e="f"/></r>', '<a b="c" d:e="f"/>', 21, 3, 'a',
             {'b': "c", 'd:e': "f"}, None, None, 21, '', 21),
            ('<r><a b="c" d="e" /></r>', '<a b="c" d="e" />', 20, 3, 'a',
             {'b': "c", 'd': "e"}, None, None, 20, '', 20),
            ('<r><a b="c/d" e="f"/></r>', '<a b="c/d" e="f"/>', 21, 3, 'a',
             {'b': "c/d", 'e': "f"}, None, None, 21, '', 21),
            ('<r><a b="c>d" e="f"/></r>', '<a b="c>d" e="f"/>', 21, 3, 'a',
             {'b': "c>d", 'e': "f"}, None, None, 21, '', 21),

            # Comments
            ('<r><!----></r>', '<!---->', 10, 3, NewDumbXml.COMMENT, {}, 7, '',
             10, '', 10),
            ('<r><!-- hello there --></r>', '<!-- hello there -->', 23, 3,
             NewDumbXml.COMMENT, {}, 7, ' hello there ', 23, '', 23),
            ('<r><!-- hello there --> tail</r>', '<!-- hello there --> tail',
             28, 3, NewDumbXml.COMMENT, {}, 7, ' hello there ', 23, ' tail',
             28),

            # With content
            ('<r><a></a></r>', '<a></a>', 10, 3, 'a', {}, 6, '', 10, '', 10),
            ('<r><a></a> tail</r>', '<a></a> tail', 15, 3, 'a', {}, 6, '', 10,
             ' tail', 15),
            ('<r><a b="c"></a> tail</r>', '<a b="c"></a> tail', 21, 3, 'a',
             {'b': "c"}, 12, '', 16, ' tail', 21),
            ('<r><a>hello world</a></r>', '<a>hello world</a>', 21, 3, 'a', {},
             6, 'hello world', 21, '', 21),
            ('<r><a>hello <![CDATA[</a>]]></a></r>',
             '<a>hello <![CDATA[</a>]]></a>', 32, 3, 'a', {}, 6,
             'hello <![CDATA[</a>]]>', 32, '', 32),
            ('<r><a>hello world</a> tail</r>', '<a>hello world</a> tail', 26,
             3, 'a', {}, 6, 'hello world', 21, ' tail', 26),
            ('<r><a b="c">hello world</a> tail</r>',
             '<a b="c">hello world</a> tail', 32, 3, 'a', {'b': "c"}, 12,
             'hello world', 27, ' tail', 32),
        )
        for case in cases:
            root = NewDumbXml(case[0])
            children = list(root)
            self.assertEquals(len(children), 1)
            inner = children[0]
            self.assertEquals(
                (inner.source, root.content, root.content_end, inner.position,
                 inner.tag, inner.attrib, inner.text_position, inner.text,
                 inner.tail_position, inner.tail, inner.end),
                case
            )

    def test_several_children(self):
        cases = (
            ['<a />', None],
            ['<a><b/><c>hello world</c></a >', '<b/><c>hello world</c>',
             '<b/>', '<c>hello world</c>'],
            ['<a><b/><c>hello world</c></a>', '<b/><c>hello world</c>', '<b/>',
             '<c>hello world</c>'],
            ['<a><b c="d"/><e>hello world</e></a>',
             '<b c="d"/><e>hello world</e>', '<b c="d"/>',
             '<e>hello world</e>'],
            ['<a><b/><c d="e">hello world</c></a>',
             '<b/><c d="e">hello world</c>', '<b/>',
             '<c d="e">hello world</c>'],
            ['<a><b><c /></b></a>', '<b><c /></b>', '<b><c /></b>'],
            ['<a><b><c /></b><d /></a>', '<b><c /></b><d />', '<b><c /></b>',
             '<d />'],
        )
        for case in cases:
            root_source = case[0]
            content = case[1]
            children_sources = case[2:]
            root = NewDumbXml(root_source)
            self.assertEquals(root.content, content)
            self.assertEquals(
                [inner.source[inner.position:inner.end] for inner in root],
                children_sources
            )

    def test_iter_errors(self):
        cases = (
            ('<a></b>',
             "Closing tag 'b' does not match opening tag 'a' on line 1"),
            ('<a></a', "Invalid closing of tag 'a' on line 1"),
            ('<a></a aosdjfio', "Invalid closing of tag 'a' on line 1"),
        )
        for source, error_msg in cases:
            with self.assertRaisesRegexp(DumbXmlSyntaxError,
                                         r'^{}$'.format(re.escape(error_msg))):
                list(NewDumbXml(source))

    def test_inbetweens(self):
        root = NewDumbXml('<a>This<b/>is<c/>separated<d/>by<e/>tags</a>')
        collected = [root.text]
        for inner in root:
            collected.append(inner.tail)
        self.assertEquals(root.content,
                          'This<b/>is<c/>separated<d/>by<e/>tags')
        self.assertEquals(collected, ['This', 'is', 'separated', 'by', 'tags'])

    def test_find_children(self):
        root = NewDumbXml("""
            <ul>
                <li>one</li>
                <something_else>something else</something_else>
                <li>two</li>
                <div>
                    <li>This won't be collected</li>
                </div>
                <li highlight="true">three</li>
            </ul>
        """)
        self.assertEquals(
            [(inner.tag, inner.attrib, inner.text)
             for inner in root.find_children('li')],
            [('li', {}, "one"), ('li', {}, "two"),
             ('li', {'highlight': "true"}, "three")]
        )

    def test_find_children_multitag(self):
        root = NewDumbXml("""
            <root>
                <a>AAA</a>
                <b>BBB</b>
                <c>CCC</c>
            </root>
        """)
        self.assertEquals(
            [(inner.tag, inner.attrib, inner.text)
             for inner in root.find_children('a', 'b')],
            [('a', {}, "AAA"),
             ('b', {}, "BBB")]
        )

    def test_find_all_children(self):
        root = NewDumbXml("""
            <ul>
                <li>one</li>
                <something_else>something else</something_else>
                <li>two</li>
                <div>
                    <li>This won't be collected</li>
                </div>
                <li highlight="true">three</li>
            </ul>
        """)
        self.assertEquals(
            [(inner.tag, inner.attrib, inner.text.strip())
             for inner in root.find_children()],
            [('li', {}, "one"),
             ('something_else', {}, "something else"),
             ('li', {}, "two"),
             ('div', {}, ""),
             ('li', {'highlight': "true"}, "three")]
        )

    def test_find_descendants(self):
        root = NewDumbXml("""
            <div>
                <p>Header</p>
                <div>
                    <p class="icon icon-subheader">Subheader</p>
                    <img src="foo" alt="bar" />
                </div>
            </div>
        """)
        self.assertEquals(
            [(inner.tag, inner.attrib, inner.text)
             for inner in root.find_descendants('p')],
            [('p', {}, "Header"),
             ('p', {'class': "icon icon-subheader"}, "Subheader")]
        )

    def test_find_descendants_multitag(self):
        root = NewDumbXml("""
            <root>
                <a>first child</a>
                <b>
                    <a>first grandchild</a>
                    <c>second grandchild</c>
                </b>
            </root>
        """)
        self.assertEquals(
            [(inner.tag, inner.attrib, inner.text)
             for inner in root.find_descendants('a', 'c')],
            [('a', {}, "first child"),
             ('a', {}, "first grandchild"),
             ('c', {}, "second grandchild")]
        )

    def test_find_all_descendants(self):
        root = NewDumbXml("""
            <div>
                <p>Header</p>
                <div>
                    <p class="icon icon-subheader">Subheader</p>
                    <img src="foo" alt="bar" />
                </div>
            </div>
        """)
        self.assertEquals(
            [(inner.tag, inner.attrib, inner.text and inner.text.strip())
             for inner in root.find_descendants()],
            [('p', {}, "Header"),
             ('div', {}, ""),
             ('p', {'class': "icon icon-subheader"}, "Subheader"),
             ('img', {'src': "foo", 'alt': "bar"}, None)]
        )



from string import ascii_letters, digits
from random import choice


def generate_random_string(length=20):
    return ''.join(choice(ascii_letters + digits) for _ in range(length))


def strip_leading_spaces(source):
    r"""
    This is to help you write multilingual strings as test inputs in your
    tests without screwing up your code's syntax. Eg::

        '''
            1
            00:01:28.797 --> 00:01:30.297 X:240 Y:480
            Hello world
        '''

    will be converted to::

        '\n1\n00:01:28.797 --> 00:01:30.297 X:240 Y:480\nHello world\n'
    """

    return '\n'.join((line.lstrip() for line in source.split('\n')))



import re


class DumbXml(object):
    """
        Describes and XML tag with its contents.

        The fact that the content starts with a '<' and ends with a '>' is
        assumed.

        You can iterate over the contents with the `find` method, which
        takes a tag name, a list of tag names or nothing as arguments and
        yields
    """

    OPENING_TAG_PAT = re.compile(
        r'^\s*\<(?P<name>[^\s\n\>]+)(?P<attrs>[^\>]*)\>',
        re.DOTALL
    )
    ATTR_PAT = re.compile(r'\b(?P<key>[^=]+)="(?P<value>[^"]+)"')
    COMMENT = "!--"
    SINGLE_TAG_PAT = re.compile(r'/\s*\>$')

    def __init__(self, content):
        """
            Does some parsing and sets the following attributes to `self`:

            * content: The content of the tag, including the opening/closing
                tags
            * name: The name of the tag
            * attrs: A dictionary of all the attributes of the tag with their
                values
            * inner_offset: the place of the character where the inner content
                of the tag starts, aka the length of the opening tag
            * inner: the inner content of the tag
        """

        self.content = content

        if self.content[:4] == "<!--":
            # Special case for comment
            self.inner_offset = 4
            self.name = self.COMMENT
            self.attrs = {}
            self.inner = self.content[4:self.content.index("-->")]
            return

        opening_match = self.OPENING_TAG_PAT.search(content)
        self.inner_offset = opening_match.end()
        self.name = opening_match.groupdict()['name']
        attrs = opening_match.groupdict()['attrs']
        self.attrs = {}
        for match in self.ATTR_PAT.finditer(attrs):
            self.attrs[match.groupdict()['key']] = match.groupdict()['value']

        closing_start, closing_end = self.find_closing(0)

        self.inner = self.content[opening_match.end():closing_start]

    def find(self, tags=[]):
        if isinstance(tags, (str, unicode)):
            tags = [tags]

        if not tags:
            pat = re.compile(r'\<', re.DOTALL)
        else:
            pat = re.compile(r'\<(?:{})'.format(
                '|'.join((re.escape(tag) for tag in tags))
            ), re.DOTALL)

        for match in pat.finditer(self.content):
            if match.start() == 0 or self._is_within_comment(match):
                continue
            closing_start, closing_end = self.find_closing(match.start())
            found = DumbXml(self.content[match.start():closing_end])
            if not tags or found.name in tags:
                offset = match.start()
                yield found, offset

    def find_closing(self, start):
        # assume start is on a '<'

        if self.content[start:start + 4] == "<!--":
            # Special case for comment
            closing_start = self.content[start:].index("-->")
            return start + closing_start, start + closing_start + 3

        opening_match = self.OPENING_TAG_PAT.search(self.content[start:])

        if self.SINGLE_TAG_PAT.search(opening_match.group()):
            # Single tag, eg `<foo a="b" />`
            return start + opening_match.end(), start + opening_match.end()

        tag_name = opening_match.groupdict()['name']
        tag_pat = re.compile(r'\<(?:(?:{tag_name})|(?:/{tag_name}\>))'.
                             format(tag_name=re.escape(tag_name)))
        match_generator = tag_pat.finditer(self.content[start:])
        first_match = match_generator.next()
        assert first_match and first_match.start() == 0 and\
            first_match.group()[1] != '/'
        count = 1
        for match in match_generator:
            matched = match.group()
            if matched[1] == '/' or matched == "-->":
                # closing tag
                count -= 1
            else:
                count += 1

            if count == 0:
                return start + match.start(), start + match.end()

    def _is_within_comment(self, match):
        # Previous opening comment
        try:
            opening = match.start() -\
                self.content[match.start()::-1].index("--!<")
        except ValueError:
            opening = None
        # Previous closing comment
        try:
            closing = match.start() -\
                self.content[match.start()::-1].index(">--")
        except ValueError:
            closing = None

        if opening is not None:
            if closing is not None:
                if closing > opening:
                    return False
                else:
                    return True
            else:
                return True
        else:
            return False


if __name__ == "__main__":
    document = DumbXml(
        '<resources><string name="foo">hello world</string></resources>'
    )
    strings = document.find('string')
    string, _ = strings.next()
    print "{}: {}".format(string.attrs['name'], string.inner)


class DumbXmlSyntaxError(Exception):
    pass


class NewDumbXml(object):
    """ A utility to help process an XML string. The main focuses are:

        1. Return the exact contents of whatever's encountered, don't unescape
           anything
        2. Return the positions of stuff encountered, relative to the root
           string.

        To initialize, simply pass an XML string:

            >>> dumb_xml = DumbXml('<a key="value">b</a> tail here')


        If you want, you can pass an extra argument to identify an embedded XML
        tag within the outer one, like this:

            >>> source = '<a><b>hello world</b></a>'
            >>> inner = DumbXml(source, start=3)

        If you do this, the positions retrieved will be relative to the outer
        string:

            >>> assert inner.text_position == 6

        The properties you can retrieve from a DumbXml instance are (for the
        example, consider this string: '<a key="value">b</a> tail here'):

        - `dumb_xml.position`: The starting position of the xml tag (in our
          example: `0`)

        - `dumb_xml.tag`: the name of the tag (in our example: `'a'`)

            - If this is a namespaced tag, this utility does not bother with it
              and will happily return something like: `'xliff:g'`.

            - In case of a comment, this will return '!--' which is equal to
              DumbXml.COMMENT

        - `dumb_xml.attrib`: The attributes of the tag, if any, in dict format
          (in our example: `{'key': "value"}`)

        - `dumb_xml.text_position`: The starting point of the tag's contents
          (in our example: `15`)

            - For single tags (eg '<br />'), `text` and `text_position` will be
              None; for empty tags (eg, '<li></li>'), `text` will be `''` and
              `text_position` will be the expected value

        - `dumb_xml.text`: The text content of the tag (in our example: `'b'`)

            - If the contents begin with text but also have other xml tags, we
              follow lxml's approach: `text` is what's contained between the
              start of the tag's content and the start of the first inner tag.
              To get the rest of the text contents, you have to access `tail`
              on the inner tags (see below).

        - `for inner_tag in dumb_xml`: Iterating over a DumbXml object will
          generate all contained tags

            - The positions returned by the inner tags' properties will be
              relative to the root string.

        - `dumb_xml.content`: If the tag contains children tags, `content` is
          the whole contents in string form, otherwise it's the same as `text`
          (in our example `'b'`)

        - `dumb_xml.content_end`: the position where the contents of the tag
          end (in our example, `16`)

        - `dumb_xml.tail`: The text contained between the end of this tag and
          either the start of the next one or the end of the source string (in
          our example: `' tail here'`)

        - `dumb_xml.tail_position`: The starting position of tail

        - `dumb_xml.end`: the end of the tail (should be the start of the next
          tag, if there is one)

        Two finding methods are supported, `find_children` and
        `find_descendants`, which accept a tag name or list of tag names as
        argument. If the argument is left None, all children and descendants
        will be yielded.
    """

    class NOT_CACHED:
        "Special value for None because for some properties, None is valid"

    COMMENT = '!--'

    def __init__(self, source, start=0):
        self.source = source
        self.start = start
        self._position = self._tag = self._attrib = self._attrib_string =\
            self._text_position = self._text = self._content_end =\
            self._tail_position = self._tail = self.NOT_CACHED

        # Start with tag because if this is a comment, it will mess up with the
        # retrieving of other attributes
        self.tag

    @property
    def position(self):
        """ The starting position of the tag.

            <atag>Some text</atag>
            ^
        """

        if self._position is not self.NOT_CACHED:
            return self._position

        self._position = self._find_next_lt(self.start)
        return self._position

    @property
    def tag(self):
        """ The name of the tag.

            <atag>Some text</atag>
             ^^^^
        """

        if self._tag is not self.NOT_CACHED:
            return self._tag

        start = self.position
        end = start + len("<!--")
        if self.source[start:end] == "<!--":
            self._tag = self.COMMENT
            self._process_comment()
            return self._tag

        for ptr in xrange(self.position + 1, len(self.source)):
            candidate = self.source[ptr]
            if candidate in ('/', '>') or candidate.isspace():
                self._tag = self.source[self.position + 1:ptr]
                return self._tag
        raise DumbXmlSyntaxError(u"Opening tag not closed on line {}".
                                 format(self._find_line_number()))

    @property
    def attrib(self):
        if self._attrib is not self.NOT_CACHED:
            return self._attrib

        attrib_start_p = self.position + 1 + len(self.tag)
        in_quotes = False
        for ptr in xrange(attrib_start_p, len(self.source)):
            candidate = self.source[ptr]
            if candidate in ('"', "'"):
                in_quotes = not in_quotes
            if not in_quotes and candidate in ('/', '>'):
                attrib_end_p = ptr
                break
        else:
            raise DumbXmlSyntaxError(
                u"Opening tag '{}' not closed on line {}".
                format(self.tag, self._find_line_number())
            )

        self._attrib_string = self.source[attrib_start_p:attrib_end_p]
        pat = re.compile(r"""(?P<key>[^\s=]+)
                             \s*=\s*
                             (
                                 ('(?P<value_apos>[^']+)') |
                                 ("(?P<value_quot>[^"]+)")
                             )""", re.VERBOSE)

        self._attrib = {}
        for match in pat.finditer(self._attrib_string):
            groupdict = match.groupdict()
            key = groupdict['key']
            value = groupdict['value_apos'] or groupdict['value_quot']
            self._attrib[key] = value
        return self._attrib

    @property
    def text_position(self):
        """ The start position of the text.

            <atag>Some text</atag>
                  ^
        """

        if self._text_position is not self.NOT_CACHED:
            return self._text_position

        if self._attrib_string is self.NOT_CACHED:
            self.attrib  # This will generate self._attrib_string

        ptr = self.position + 1 + len(self.tag) + len(self._attrib_string)
        candidate = self.source[ptr]
        # Based on how we calculated '_attrib_string', this should either be
        # '/' or '>'
        if candidate == '/':  # This is a "single-tag", eg '<br />'
            self._text_position = None
            start = ptr + 1
            for ptr in xrange(start, len(self.source)):
                candidate = self.source[ptr]
                if candidate.isspace():
                    continue
                elif candidate == '>':
                    self._tail_position = ptr + 1
                    return self._text_position
                else:
                    raise DumbXmlSyntaxError(
                        u"Opening tag '{}' not closed on line {}".
                        format(self.tag, self._find_line_number())
                    )
            raise DumbXmlSyntaxError(
                u"Opening tag '{}' not closed on line {}".
                format(self.tag, self._find_line_number())
            )
        elif candidate == '>':
            self._text_position = ptr + 1
            return self._text_position
        else:
            raise DumbXmlSyntaxError(u"Something went wrong")

    @property
    def text(self):
        """ The text of the tag (up until the first child tag).

            <atag>Some <b>text</b></atag>
                  ^^^^^
        """

        if self._text is not self.NOT_CACHED:
            return self._text

        if self.text_position is None:
            self._text = None
            return self._text

        next_tag_position = self._find_next_lt(self.text_position)
        if next_tag_position == len(self.source):
            raise DumbXmlSyntaxError(
                u"Tag '{}' not closed on line {}".
                format(self.tag, self._find_line_number())
            )

        self._text = self.source[self.text_position:next_tag_position]
        return self._text

    def __iter__(self):
        if self.text is None or self.tag == self.COMMENT:
            return

        start = self.text_position + len(self.text)
        while True:
            if self.source[start + 1] == '/':  # We found the closing tag
                self._content_end = start
                closing_tag = self.source[start + 2: start + 2 + len(self.tag)]
                if closing_tag != self.tag:
                    raise DumbXmlSyntaxError(
                        u"Closing tag '{}' does not match opening tag '{}' on "
                        u"line {}".
                        format(closing_tag, self.tag, self._find_line_number())
                    )
                for ptr in xrange(start + 2 + len(self.tag), len(self.source)):
                    candidate = self.source[ptr]
                    if candidate.isspace():
                        continue
                    elif candidate == '>':
                        self._tail_position = ptr + 1
                        return
                    else:
                        raise DumbXmlSyntaxError(
                            u"Invalid closing of tag '{}' on line {}".
                            format(self.tag, self._find_line_number())
                        )
                raise DumbXmlSyntaxError(
                    u"Invalid closing of tag '{}' on line {}".
                    format(self.tag, self._find_line_number())
                )
            else:
                # Use `self.__class__` in case this is a subclass (eg to handle
                # HTML)
                inner = self.__class__(self.source, start)
                yield inner
                start = inner.end

    @property
    def content_end(self):
        """ The end of all contents of a tag (both text and children tags)

            <a>goobye <b>cruel</b> world</a>
                                        ^
        """

        if self._content_end is not self.NOT_CACHED:
            return self._content_end
        if self.text_position is None:
            return None
        for _ in self:
            pass
        return self._content_end

    @property
    def content(self):
        """ All the contents of a tag (both text and children tags)

            <a>goobye <b>cruel</b> world</a>
               ^^^^^^^^^^^^^^^^^^^^^^^^^
        """

        if self.tag == self.COMMENT:
            return self.text
        if self.content_end is None:
            return None
        return self.source[self.text_position:self.content_end]

    @property
    def tail_position(self):
        """ The position just after the tag.

            <atag>Some text</atag> newlines etc <anothertag>...
                                  ^
        """

        if self._tail_position is not self.NOT_CACHED:
            return self._tail_position

        # _tail_position might be set when getting 'text_position'
        self.text_position
        if self._tail_position is not self.NOT_CACHED:
            return self._tail_position

        # Otherwise, it should be set after iterating
        for _ in self:
            pass
        return self._tail_position

    @property
    def tail(self):
        """The text that follows the tag untill the start of a news one.

            <atag>Some text</atag> newlines etc <anothertag>...
                                  ^^^^^^^^^^^^^^
        """

        if self._tail is not self.NOT_CACHED:
            return self._tail

        self._tail = self.source[self.tail_position:
                                 self._find_next_lt(self.tail_position)]
        return self.tail

    @property
    def end(self):
        """The starting position of the next tag.

            <atag>Some text</atag> newlines etc <anothertag>...
                                                ^
        """

        return self.tail_position + len(self.tail)

    def find_children(self, *tags):
        for child in self:
            if not tags or child.tag in tags:
                yield child

    def find_descendants(self, *tags):
        for child in self:
            if not tags or child.tag in tags:
                yield child
            for inner in child.find_descendants(*tags):
                yield inner

    def _find_next_lt(self, start):
        in_cdata = False
        for ptr in xrange(start, len(self.source)):
            candidate = self.source[ptr]
            if in_cdata:
                if (candidate == ']' and
                        self.source[ptr:ptr + len("]]>")] == "]]>"):
                    in_cdata = False
            else:
                if candidate == "<":
                    # Check against CDATA
                    if self.source[ptr:ptr + len("<![CDATA[")] == "<![CDATA[":
                        in_cdata = True
                    else:
                        return ptr
        # We reached the end of the string, lets return accordingly
        return len(self.source)

    def _process_comment(self):
        # We already know position and tag

        self._attrib_string = ""
        self._attrib = {}

        self._text_position = self.position + len("<!--")
        for ptr in xrange(self._text_position, len(self.source)):
            candidate = self.source[ptr]
            if candidate == '-' and self.source[ptr:ptr + len("-->")] == "-->":
                self._content_end = ptr
                self._text = self.source[self.text_position:ptr]
                self._tail_position = ptr + len("-->")
                return
        raise DumbXmlSyntaxError(u"Comment not closed on line {}".
                                 format(self._find_line_number()))

    def _find_line_number(self, ptr=None):
        ptr = ptr or self.position
        return self.source[:ptr].count('\n') + 1



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import json
import re
from itertools import count

from ..exceptions import ParseError
from ..handlers import Handler
from ..strings import OpenString
from ..transcribers import Transcriber
from ..utils.json import DumbJson


class JsonHandler(Handler):
    name = "KEYVALUEJSON"
    extension = "json"

    def parse(self, content):
        # Do a first pass using the `json` module to ensure content is valid
        try:
            json.loads(content)
        except ValueError, e:
            raise ParseError(e.message)

        self.transcriber = Transcriber(content)
        source = self.transcriber.source
        self.stringset = []
        self.existing_keys = set()

        try:
            parsed = DumbJson(source)
        except ValueError, e:
            raise ParseError(e.message)
        self._order = count()
        self._extract(parsed)
        self.transcriber.copy_until(len(source))

        return self.transcriber.get_destination(), self.stringset

    def _extract(self, parsed, nest=None):
        if parsed.type == dict:
            for key, key_position, value, value_position in parsed:
                key = self._escape_key(key)
                if nest is not None:
                    key = u"{}.{}".format(nest, key)

                # 'key' should be unique
                if key in self.existing_keys:
                    # Need this for line number
                    self.transcriber.copy_until(key_position)
                    raise ParseError(u"Duplicate string key ('{}') in line {}".
                                     format(key, self.transcriber.line_number))
                self.existing_keys.add(key)

                if isinstance(value, (str, unicode)):
                    if not value.strip():
                        continue
                    openstring = OpenString(key, value,
                                            order=next(self._order))
                    self.transcriber.copy_until(value_position)
                    self.transcriber.add(openstring.template_replacement)
                    self.transcriber.skip(len(value))
                    self.stringset.append(openstring)
                elif isinstance(value, DumbJson):
                    self._extract(value, key)
                else:
                    # Ignore other JSON types (bools, nulls, numbers)
                    pass
        elif parsed.type == list:
            for index, (item, item_position) in enumerate(parsed):
                if nest is None:
                    key = u"..{}..".format(index)
                else:
                    key = u"{}..{}..".format(nest, index)
                if isinstance(item, (str, unicode)):
                    if not item.strip():
                        continue
                    openstring = OpenString(key, item, order=next(self._order))
                    self.transcriber.copy_until(item_position)
                    self.transcriber.add(openstring.template_replacement)
                    self.transcriber.skip(len(item))
                    self.stringset.append(openstring)
                elif isinstance(item, DumbJson):
                    self._extract(item, key)
                else:
                    # Ignore other JSON types (bools, nulls, numbers)
                    pass
        else:
            raise ParseError("Invalid JSON")

    @staticmethod
    def _escape_key(key):
        key = key.replace(u"\\", u"\\\\")
        key = key.replace(u".", u"\\.")
        return key

    def compile(self, template, stringset):
        # Lets play on the template first, we need it to not include the hashes
        # that aren't in the stringset. For that we will create a new stringset
        # which will have the hashes themselves as strings and compile against
        # that. The compilation process will remove any string sections that
        # are absent from the stringset. Next we will call `_clean_empties`
        # from the template to clear out any `...,  ,...` or `...{ ,...`
        # sequences left. The result will be used as the actual template for
        # the compilation process

        stringset = list(stringset)

        fake_stringset = [OpenString(openstring.key,
                                     openstring.template_replacement,
                                     order=openstring.order)
                          for openstring in stringset]
        new_template = self._replace_translations(template, fake_stringset)
        new_template = self._clean_empties(new_template)

        return self._replace_translations(new_template, stringset)

    def _replace_translations(self, template, stringset):
        self.transcriber = Transcriber(template)
        template = self.transcriber.source
        self.stringset = stringset
        self.stringset_index = 0

        parsed = DumbJson(template)
        self._intract(parsed)

        self.transcriber.copy_until(len(template))
        return self.transcriber.get_destination()

    def _intract(self, parsed):
        if parsed.type == dict:
            return self._intract_dict(parsed)
        elif parsed.type == list:
            return self._intract_list(parsed)

    def _intract_dict(self, parsed):
        at_least_one = False
        for key, key_position, value, value_position in parsed:
            self.transcriber.copy_until(key_position - 1)
            self.transcriber.mark_section_start()
            if isinstance(value, (str, unicode)):
                string = self._get_next_string()
                if (string is not None and
                        value == string.template_replacement):
                    at_least_one = True
                    self.transcriber.copy_until(value_position)
                    self.transcriber.add(string.string)
                    self.transcriber.skip(len(value))
                    self.stringset_index += 1
                else:
                    self.transcriber.copy_until(value_position +
                                                len(value) + 1)
                    self.transcriber.mark_section_end()
                    self.transcriber.remove_section()
            elif isinstance(value, DumbJson):
                all_removed = self._intract(value)
                if all_removed:
                    self.transcriber.copy_until(value.end + 1)
                    self.transcriber.mark_section_end()
                    self.transcriber.remove_section()
                else:
                    at_least_one = True
            else:
                # 'value' is a python value allowed by JSON (integer,
                # boolean, null), skip it
                at_least_one = True
        return not at_least_one

    def _intract_list(self, parsed):
        at_least_one = False
        for item, item_position in parsed:
            self.transcriber.copy_until(item_position - 1)
            self.transcriber.mark_section_start()
            if isinstance(item, (str, unicode)):
                string = self._get_next_string()
                if (string is not None and
                        item == string.template_replacement):
                    at_least_one = True
                    self.transcriber.copy_until(item_position)
                    self.transcriber.add(string.string)
                    self.transcriber.skip(len(item))
                    self.stringset_index += 1
                else:
                    self.transcriber.copy_until(item_position + len(item) + 1)
                    self.transcriber.mark_section_end()
                    self.transcriber.remove_section()
            elif isinstance(item, DumbJson):
                all_removed = self._intract(item)
                if all_removed:
                    self.transcriber.copy_until(item.end + 1)
                    self.transcriber.mark_section_end()
                    self.transcriber.remove_section()
                else:
                    at_least_one = True
            else:
                # 'value' is a python value allowed by JSON (integer,
                # boolean, null), skip it
                at_least_one = True
        return not at_least_one

    def _clean_empties(self, compiled):
        """ If sections were removed, clean leftover commas, brackets etc.

            Eg:
                '{"a": "b", ,"c": "d"}' -> '{"a": "b", "c": "d"}'
                '{, "a": "b", "c": "d"}' -> '{"a": "b", "c": "d"}'
                '["a", , "b"]' -> '["a", "b"]'
        """
        while True:
            # First key-value of a dict was removed
            match = re.search(r'{\s*,', compiled)
            if match:
                compiled = u"{}{{{}".format(compiled[:match.start()],
                                            compiled[match.end():])
                continue

            # Last key-value of a dict was removed
            match = re.search(r',\s*}', compiled)
            if match:
                compiled = u"{}}}{}".format(compiled[:match.start()],
                                            compiled[match.end():])
                continue

            # First item of a list was removed
            match = re.search(r'\[\s*,', compiled)
            if match:
                compiled = u"{}[{}".format(compiled[:match.start()],
                                           compiled[match.end():])
                continue

            # Last item of a list was removed
            match = re.search(r',\s*\]', compiled)
            if match:
                compiled = u"{}]{}".format(compiled[:match.start()],
                                           compiled[match.end():])
                continue

            # Intermediate key-value of a dict or list was removed
            match = re.search(r',\s*,', compiled)
            if match:
                compiled = u"{},{}".format(compiled[:match.start()],
                                           compiled[match.end():])
                continue

            # No substitutions happened, break
            break

        return compiled

    def _get_next_string(self):
        try:
            return self.stringset[self.stringset_index]
        except IndexError:
            return None

    @staticmethod
    def escape(string):
        escaped_string = string.replace('\\', r'\\').replace('"', '\\"')
        return escaped_string

    @staticmethod
    def unescape(string):
        unescaped_string = string.replace(r'\\', '\\').replace(r'\"', '"')
        return unescaped_string



from hashlib import md5


class OpenString(object):
    """
    This class will abstract away the business of generating hashes out of your
    strings and will serve as a place to get translations from when compiling.
    Several OpenStrings in our process define a *Stringset*, which is simply a
    python list of OpenStrings. To create an OpenString, you need 2 arguments:

    * The 'key'

      Something in your source file that uniquely identifies the section that
      the source string originated from. It might be helpful for your
      compiler to use something that appears in the same form in language
      files as well.

    * The 'string' or 'plural forms of the string':

      If the file format you're working with does not support plural forms,
      or if the string in question is not pluralized, you can just supply the
      string itself as the second argument. If you string is pluralized
      however, you have to supply all plural forms in a dictionary with the
      rule numbers as keys. For example::

          OpenString("UNREAD MESSAGES",
                     {1: "You have %s unread message",
                      5: "You have %s unread messages"})

    * There are a number of optional keyword arguments to `OpenString`:

      context, order, character_limit, occurrences, developer_comment, flags,
      fuzzy, obsolete

    Their main purpose is to provide context to the translators so that they
    can achieve higher quality. Two of them however, though optional, are
    highly recommended:

    * Context

      This is also taken into account when producing the hash, so if you
      can't ensure that your keys aren't unique within the source file, you
      can still get away with ensuring that the `(key, context)` pair is.

    * Order

      If you provide an order (integer), Transifex will save it in the
      database and then, when you try to compile a template against a
      stringset fetched from Transifex, it will already be ordered, even if
      it contains translations. This can allow you to optimize the
      compilation process as the order that the hashes appear in the template
      will be the same as the order of strings in the stringset.

      Another valuable outcome is that the order will be preserved when the
      strings are shown to translators which can provide context and thus
      improve translation quality.

    Once you have created an OpenString, you can get it's hash using the
    `template_replacement` property
    """

    DEFAULTS = {
        'context': "",
        'order': None,
        'character_limit': None,
        'occurrences': None,
        'developer_comment': "",
        'flags': "",
        'fuzzy': False,
        'obsolete': False,
    }

    def __init__(self, key, string_or_strings, **kwargs):
        self.key = key
        if isinstance(string_or_strings, dict):
            self.pluralized = len(string_or_strings) > 1
            self._strings = {key: value
                             for key, value in string_or_strings.items()}
        else:
            self.pluralized = False
            self._strings = {5: string_or_strings}

        for key, value in self.DEFAULTS.items():
            setattr(self, key, kwargs.get(key, value))

        if 'pluralized' in kwargs:
            self.pluralized = kwargs['pluralized']

        self._string_hash = None

    def __hash__(self):
        return hash((self.key, self.context, self.rule))

    def __repr__(self):
        return '"{}"'.format(self._strings[5].encode('utf-8'))  # pragma: nocover # noqa

    @property
    def string(self):
        if self.pluralized:
            return self._strings
        else:
            return self._strings[5]

    def _get_string_hash(self):
        keys = [self.key, self.context or '']
        return md5(':'.join(keys).encode('utf-8')).hexdigest()

    @property
    def string_hash(self):
        if self._string_hash is None:
            self._string_hash = self._get_string_hash()
        return self._string_hash

    @property
    def template_replacement(self):
        suffix = 'pl' if self.pluralized else 'tr'
        return "{hash}_{suffix}".format(hash=self.string_hash, suffix=suffix)



class Transcriber(object):
    r"""
    This class helps with creating a template from an imported file or compile
    an output file from a template.

    **Main functionality**

    This class will help with both creating a template from an imported file
    and with compiling a file from a template. It provides functions for
    copying text. It depends on 3 things, the source content (self.source), the
    target content (self.destination) which initially will contain an empty
    string and a pointer (self.ptr) which will indicate which parts of 'source'
    have already been copied to 'destination' (and will be initialized to 0).

    Transcriber detects and remembers the newline type (DOS, ``'\r\n'`` or UNIX
    ``'\n'``) of 'source'. It then converts 'source' to UNIX-like newlines and
    works on this. When returning the destination, the initial newline type
    will be used. Because 'source' is being potentially edited, it's a good
    idea to save Transcriber's source back on top of the original one:

        >>> def parse(self, source):
        ...     self.transcriber = Transcriber(source)
        ...     self.source = self.transcriber.source
        ...     # ...

    The main methods provided are demonstrated below::

        >>> transcriber = Transcriber(source)

        source:      <string name="foo">hello world</string>
        ptr:         ^ (0)
        destination: []

        >>> transcriber.copy_until(source.index('>') + 1)

        source:      <string name="foo">hello world</string>
        ptr:                            ^
        destination: ['<string name="foo">']

        >>> transcriber.add("aee8cc2abd5abd5a87cd784be_tr")

        source:      <string name="foo">hello world</string>
        ptr:                            ^
        destination: ['<string name="foo">', 'aee8cc2abd5abd5a87cd784be_tr']

        >>> transcriber.skip(len("hello world"))

        source:      <string name="foo">hello world</string>
        ptr:                                       ^
        destination: ['<string name="foo">', 'aee8cc2abd5abd5a87cd784be_tr']

        >>> transcriber.copy_until(source.index("</string>") +
        ...                        len("</string>"))

        source:      <string name="foo">hello world</string>
        ptr:                                                ^
        destination: ['<string name="foo">', 'aee8cc2abd5abd5a87cd784be_tr',
        '</string>']

        >>> print transcriber.get_destination()

        <string name="foo">aee8cc2abd5abd5a87cd784be_tr</string>
    """

    class SectionStart:
        pass

    class SectionEnd:
        pass

    def __init__(self, source):
        self.source = source
        self.destination = []
        self.ptr = 0

        self.newline_count = 0

        # Handle newlines
        self.newline_type = "UNIX"
        if '\r\n' in self.source:
            self.newline_type = "DOS"
            self.source = self.source.replace('\r\n', '\n')

    def copy(self, offset):
        chunk = self.source[self.ptr:self.ptr + offset]
        self.destination.append(chunk)
        self.ptr += offset

        self.newline_count += chunk.count('\n')

    def copy_until(self, end):
        chunk = self.source[self.ptr:end]
        self.destination.append(chunk)
        self.ptr = end

        self.newline_count += chunk.count('\n')

    def add(self, text):
        self.destination.append(text)

    def skip(self, offset):
        chunk = self.source[self.ptr:self.ptr + offset]
        self.newline_count += chunk.count('\n')

        self.ptr += offset

    def skip_until(self, end):
        chunk = self.source[self.ptr:end]
        self.newline_count += chunk.count('\n')

        self.ptr = end

    def mark_section_start(self):
        self.destination.append(self.SectionStart)

    def mark_section_end(self):
        self.destination.append(self.SectionEnd)

    def remove_section(self, place=0):
        """
        You can mark sections in the target file and optionally remove them.
        Insert the section-start and section-end bookmarks wherever you want to
        mark a section. Then you can remove a section with `remove_section()`.
        For example::

            >>> transcriber = Transcriber(source)

            source:      <keep><remove>
            ptr:         ^ (0)
            destination: []

            >>> start = 0

            >>> transcriber.mark_section_start()
            >>> transcriber.copy_until(start + 1)  # copy until first '<'
            >>> string = source[start + 1:source.index('>', start)]
            >>> transcriber.add("asdf")  # add the hash
            >>> transcriber.skip(len(string))
            >>> transcriber.copy_until(source.index('>', start) + 1)
            >>> transcriber.mark_section_end()

            source:      <keep><remove>
            ptr:               ^
            destination: [SectionStart, '<', 'asdf', '>', SectionEnd]

            >>> if string == "remove":
            ...     transcriber.remove_section()

            (nothing happens)

            >>> start = source.index('>') + 1

            >>> # Same deal as before, mostly
            >>> transcriber.mark_section_start()
            >>> transcriber.copy_until(start + 1)  # copy until second '<'
            >>> string = source[start + 1:source.index('>', start)]
            >>> transcriber.add("fdsa")  # add the hash
            >>> transcriber.skip(len(string))
            >>> transcriber.copy_until(source.index('>', start) + 1)
            >>> transcriber.mark_section_end()

            source:      <keep><remove>
            ptr:                       ^
            destination: [SectionStart, '<', 'asdf', '>', SectionEnd,
                          SectionStart, '<', 'fdsa', '>', SectionEnd]

            >>> if string == "remove":
            ...     transcriber.remove_section()

            source:      <keep><remove>
            ptr:                       ^
            destination: [SectionStart,  '<', 'asdf', '>',  SectionEnd,
                          None        , None, None  , None, None      ]

            (The last section was replaced with Nones)

            Now, when you try to get the result with `get_destination()`, the
            Nones, SectionStarts and SectionEnds will be ommited:

            >>> transcriber.get_destination()

            <asdf>
        """
        section_start_position = self._find_last_section_start(place)
        try:
            section_end_position = self.destination.index(
                self.SectionEnd, section_start_position
            )
        except ValueError:
            section_end_position = len(self.destination) - 1
        for i in range(section_start_position, section_end_position + 1):
            self.destination[i] = None

    def _find_last_section_start(self, place=0):
        count = place
        for i, segment in enumerate(self.destination[::-1], start=1):
            if segment == self.SectionStart:
                if count == 0:
                    return len(self.destination) - i
                else:
                    count -= 1

    @property
    def line_number(self):
        r"""
        The transcriber remembers how many newlines it has went over on the
        source, both when copying and skipping content. This allows you to
        pinpoint the line-number a parse-error has occured. For example::

            source:
                first line
                second line
                third line with error
                fourth line

            >>> transcriber = Transcriber(source)
            >>> for line in source.split("\n"):
            >>>     if "error" not in line:
            >>>         # include the newline too
            >>>         transcriber.copy(len(line) + 1)
            >>>     else:
            >>>         raise ParseError(
            >>>             "Error on line {line_no}: '{line}'".format(
            >>>                 line_no=transcriber.line_number,
            >>>                 line=line
            >>>             )
            >>>         )

            This will raise a::

            >>> ParseError("Error on line 3: 'third line with error'")
        """
        return self.newline_count + 1

    def get_destination(self, enforce_newline_type=None):
        return "".join([self.edit_newlines(chunk, enforce_newline_type)
                        for chunk in self.destination
                        if chunk not in (self.SectionStart, self.SectionEnd,
                                         None)])

    def edit_newlines(self, chunk, enforce_newline_type=None):
        r"""
        This is the part that renders the newlines to their correct type when
        returning the final result. You have the option to enforce the newline
        type if you want to.

            >>> source = "hello\r\nworld"
            >>> t = Transcriber(source)
            >>> t.source

            >>> "hello\nworld"

            >>> source = trascriber.source
            >>> # Work as if source was UNIX-type
            >>> t.copy_until(source.index('\n') + 1)  # include the '\n'
            >>> t.add("fellas")
            >>> t.get_destination()

            >>> "hello\r\nfellas"  # <- it remembered newline type from source

            >>> t.get_destination(enforce_newline_type="UNIX")

            >>> "hello\nfellas"
        """

        if ((enforce_newline_type is None and self.newline_type == "DOS") or
                enforce_newline_type == "DOS"):
            return chunk.replace('\n', '\r\n')
        else:
            return chunk



from django.conf.urls import patterns, url
from django.views.decorators.csrf import csrf_exempt

from testbed.main.views import MainView, ApiView, SaveView


urlpatterns = patterns('',  # noqa
    url(r'^$', MainView.as_view(), name="testbed_home"),
    url(r'^(?P<payload_hash>\w{32})$', MainView.as_view(),
        name="testbed_main"),
    url(r'^api/$', csrf_exempt(ApiView.as_view()), name="testbed_api"),
    url(r'^save/$', SaveView.as_view(), name="testbed_save")
)



# -*- coding: utf-8 -*-

import unittest

from openformats.strings import OpenString

from openformats.tests.formats.common import CommonFormatTestMixin
from openformats.tests.utils.strings import generate_random_string

from openformats.formats.json import JsonHandler


class JsonTestCase(CommonFormatTestMixin, unittest.TestCase):
    HANDLER_CLASS = JsonHandler
    TESTFILE_BASE = "openformats/tests/formats/beta_keyvaluejson/files"

    def setUp(self):
        super(JsonTestCase, self).setUp()

        self.handler = JsonHandler()
        self.random_string = generate_random_string()
        self.random_openstring = OpenString("a", self.random_string, order=0)
        self.random_hash = self.random_openstring.template_replacement

    def test_simple(self):
        # Using old-timey string formatting because of conflicts with '{'
        template, stringset = self.handler.parse('{"a": "%s"}' %
                                                 self.random_string)
        compiled = self.handler.compile(template, [self.random_openstring])

        self.assertEquals(template, '{"a": "%s"}' % self.random_hash)
        self.assertEquals(len(stringset), 1)
        self.assertEquals(stringset[0].__dict__,
                          self.random_openstring.__dict__)
        self.assertEquals(compiled, '{"a": "%s"}' % self.random_string)

    def test_empty_string_ignored(self):
        template, stringset = self.handler.parse(
            '{"not_empty": "hello there", "empty": ""}'
        )
        self.assertEquals(len(stringset), 1)

        template, stringset = self.handler.parse(
            '{"not_empty": "hello there", "empty": [""]}'
        )
        self.assertEquals(len(stringset), 1)

    def test_root_object_is_list(self):
        source = '["%s"]' % self.random_string
        random_openstring = OpenString('..0..', self.random_string, order=0)
        random_hash = random_openstring.template_replacement

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [random_openstring])

        self.assertEquals(template, '["%s"]' % random_hash)
        self.assertEquals(len(stringset), 1)
        self.assertEquals(stringset[0].__dict__, random_openstring.__dict__)
        self.assertEquals(compiled, source)

    def test_embedded_dicts(self):
        source = '{"a": {"b": "%s"}}' % self.random_string
        openstring = OpenString("a.b", self.random_string, order=0)
        random_hash = openstring.template_replacement

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring])

        self.assertEquals(template, '{"a": {"b": "%s"}}' % random_hash)
        self.assertEquals(len(stringset), 1)
        self.assertEquals(stringset[0].__dict__, openstring.__dict__)
        self.assertEquals(compiled, source)

    def test_embedded_lists(self):
        source = '{"a": ["%s"]}' % self.random_string
        openstring = OpenString("a..0..", self.random_string, order=0)
        random_hash = openstring.template_replacement

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring])

        self.assertEquals(template, '{"a": ["%s"]}' % random_hash)
        self.assertEquals(len(stringset), 1)
        self.assertEquals(stringset[0].__dict__, openstring.__dict__)
        self.assertEquals(compiled, source)

    def test_python_values_are_ignored(self):
        source = '[true, "%s", 5e12]' % self.random_string
        random_openstring = OpenString('..1..', self.random_string, order=0)
        random_hash = random_openstring.template_replacement
        template, stringset = self.handler.parse(source)
        self.assertEquals(template, '[true, "%s", 5e12]' % random_hash)
        self.assertEquals(len(stringset), 1)
        self.assertEquals(stringset[0].__dict__, random_openstring.__dict__)

    def test_compile_skips_removed_strings_for_dicts(self):
        string1 = self.random_string
        string2 = generate_random_string()
        openstring1 = self.random_openstring
        openstring2 = OpenString("b", string2, order=1)
        hash1 = self.random_hash
        hash2 = openstring2.template_replacement

        source = '{"a": "%s", "b": "%s"}' % (string1, string2)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring1])

        self.assertEquals(template, '{"a": "%s", "b": "%s"}' % (hash1, hash2))
        self.assertEquals(len(stringset), 2)
        self.assertEquals(stringset[0].__dict__, openstring1.__dict__)
        self.assertEquals(stringset[1].__dict__, openstring2.__dict__)
        self.assertEquals(compiled, '{"a": "%s"}' % string1)

    def test_compile_skips_removed_strings_for_lists(self):
        string1 = self.random_string
        string2 = generate_random_string()
        openstring1 = OpenString("..0..", string1, order=0)
        openstring2 = OpenString("..1..", string2, order=1)
        hash1 = openstring1.template_replacement
        hash2 = openstring2.template_replacement

        source = '["%s", "%s"]' % (string1, string2)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring1])

        self.assertEquals(template, '["%s", "%s"]' % (hash1, hash2))
        self.assertEquals(len(stringset), 2)
        self.assertEquals(stringset[0].__dict__, openstring1.__dict__)
        self.assertEquals(stringset[1].__dict__, openstring2.__dict__)
        self.assertEquals(compiled, '["%s"]' % string1)

    def test_compile_skips_removed_nested_dict(self):
        string1 = self.random_string
        string2 = generate_random_string()
        openstring1 = OpenString("..0..", string1, order=0)
        openstring2 = OpenString("..1...a", string2, order=1)
        hash1 = openstring1.template_replacement
        hash2 = openstring2.template_replacement

        source = '["%s", {"a": "%s"}]' % (string1, string2)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring1])

        self.assertEquals(template, '["%s", {"a": "%s"}]' % (hash1, hash2))
        self.assertEquals(len(stringset), 2)
        self.assertEquals(stringset[0].__dict__, openstring1.__dict__)
        self.assertEquals(stringset[1].__dict__, openstring2.__dict__)
        self.assertEquals(compiled, '["%s"]' % string1)

    def test_compile_skips_removed_nested_list(self):
        string1 = self.random_string
        string2 = generate_random_string()
        openstring1 = self.random_openstring
        openstring2 = OpenString("b..0..", string2, order=1)
        hash1 = self.random_hash
        hash2 = openstring2.template_replacement

        source = '{"a": "%s", "b": ["%s"]}' % (string1, string2)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring1])

        self.assertEquals(template,
                          '{"a": "%s", "b": ["%s"]}' % (hash1, hash2))
        self.assertEquals(len(stringset), 2)
        self.assertEquals(stringset[0].__dict__, openstring1.__dict__)
        self.assertEquals(stringset[1].__dict__, openstring2.__dict__)
        self.assertEquals(compiled, '{"a": "%s"}' % string1)

    def test_remove_all_strings_removed_from_list_but_non_strings_exist(self):
        random_string = self.random_string
        random_openstring = OpenString('a..1..', random_string, order=0)
        random_hash = random_openstring.template_replacement

        source = '{"a": [null, "%s"]}' % random_string

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [])

        self.assertEquals(template, '{"a": [null, "%s"]}' % random_hash)
        self.assertEquals(len(stringset), 1)
        self.assertEquals(stringset[0].__dict__, random_openstring.__dict__)
        self.assertEquals(compiled, '{"a": [null]}')

    def test_remove_all_strings_removed_from_dict_but_non_strings_exist(self):
        random_string = self.random_string
        random_openstring = self.random_openstring
        random_hash = self.random_hash

        source = '{"a": "%s", "b": null}' % random_string

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [])

        self.assertEquals(template, '{"a": "%s", "b": null}' % random_hash)
        self.assertEquals(len(stringset), 1)
        self.assertEquals(stringset[0].__dict__, random_openstring.__dict__)
        self.assertEquals(compiled, '{ "b": null}')

    def test_skip_start_of_dict(self):
        string1 = self.random_string
        string2 = generate_random_string()
        openstring2 = OpenString('b', string2, order=1)

        source = '{"a": "%s", "b": "%s"}' % (string1, string2)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring2])

        self.assertEquals(compiled, '{ "b": "%s"}' % string2)

    def test_skip_end_of_dict(self):
        string1 = self.random_string
        string2 = generate_random_string()
        openstring1 = OpenString('a', string1, order=1)

        source = '{"a": "%s", "b": "%s"}' % (string1, string2)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring1])

        self.assertEquals(compiled, '{"a": "%s"}' % string1)

    def test_skip_middle_of_dict(self):
        string1 = self.random_string
        string2 = generate_random_string()
        string3 = generate_random_string()
        openstring1 = OpenString('a', string1, order=0)
        openstring3 = OpenString('c', string3, order=2)

        source = '{"a": "%s", "b": "%s", "c": "%s"}' % (string1, string2,
                                                        string3)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring1, openstring3])

        self.assertEquals(compiled, '{"a": "%s", "c": "%s"}' % (string1,
                                                                string3))

    def test_skip_start_of_list(self):
        string1 = self.random_string
        string2 = generate_random_string()
        openstring2 = OpenString('..1..', string2, order=1)

        source = '["%s", "%s"]' % (string1, string2)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring2])

        self.assertEquals(compiled, '[ "%s"]' % string2)

    def test_skip_end_of_list(self):
        string1 = self.random_string
        string2 = generate_random_string()
        openstring1 = OpenString('..0..', string1, order=0)

        source = '["%s", "%s"]' % (string1, string2)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring1])

        self.assertEquals(compiled, '["%s"]' % string1)

    def test_skip_middle_of_list(self):
        string1 = self.random_string
        string2 = generate_random_string()
        string3 = generate_random_string()
        openstring1 = OpenString('..0..', string1, order=0)
        openstring3 = OpenString('..2..', string3, order=2)

        source = '["%s", "%s", "%s"]' % (string1, string2, string3)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [openstring1, openstring3])

        self.assertEquals(compiled, '["%s", "%s"]' % (string1, string3))

    def test_invalid_json(self):
        self._test_parse_error('jaosjf', "No JSON object could be decoded")

    def test_invalid_json_type(self):
        template, stringset = self.handler.parse('[false]')
        self.assertEqual(stringset, [])
        self.assertEqual(template, '[false]')

        template, stringset = self.handler.parse('{"false": false}')
        self.assertEqual(stringset, [])
        self.assertEqual(template, '{"false": false}')

    def test_not_json_container(self):
        self._test_parse_error('"hello"',
                               'Was expecting whitespace or one of `{[` on '
                               'line 1, found `"` instead')
        self._test_parse_error('3',
                               "Was expecting whitespace or one of `{[` on "
                               "line 1, found `3` instead")
        self._test_parse_error('false',
                               "Was expecting whitespace or one of `{[` on "
                               "line 1, found `f` instead")

    def test_skipping_stuff_within_strings(self):
        source = '{"a": "b,  ,c"}'
        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, stringset)
        self.assertEquals(compiled, source)

    def test_duplicate_keys(self):
        self._test_parse_error('{"a": "hello", "a": "world"}',
                               "Duplicate string key ('a') in line 1")

    def test_display_json_errors(self):
        self._test_parse_error('["]',
                               "Unterminated string starting at: line 1 "
                               "column 2 (char 1)")

    def test_escape_json(self):
        self.assertEqual(
            self.handler.escape("a \\ string. with \"quotes\""),
            "a \\\\ string. with \\\"quotes\\\""
        )

        self.assertEqual(
            self.handler.escape(u'καλημέρα \\'), u'καλημέρα \\\\'
        )

    def test_unescape_json(self):
        self.assertEqual(
            self.handler.unescape("a \\\\ string. with \\\"quotes\\\""),
            "a \\ string. with \"quotes\""
        )

        self.assertEqual(
            self.handler.unescape(u'καλημέρα \\\\'),
            u'καλημέρα \\'
        )



from random import randint
from unittest import TestCase

from openformats.exceptions import RuleError
from openformats.handlers import Handler

from openformats.tests.utils import generate_random_string


class HandlerTestCase(TestCase):
    def setUp(self):
        self.handler = Handler()

    def tearDown(self):
        self.handler = None

    def test_get_rule_number_returns(self):
        random_string = generate_random_string()
        random_integer = randint(1, 50)

        self.handler.__class__._RULES_ATOI = {
            random_string: random_integer
        }
        self.assertEqual(
            self.handler.get_rule_number(random_string), random_integer
        )

    def test_get_rule_string_returns(self):
        random_string = generate_random_string()
        random_integer = randint(1, 50)

        self.handler.__class__._RULES_ITOA = {
            random_integer: random_string
        }
        self.assertEqual(
            self.handler.get_rule_string(random_integer), random_string
        )

    def test_get_rule_string_returns_error(self):
        self.assertRaises(RuleError, self.handler.get_rule_string, 50)

    def test_get_rule_number_returns_error(self):
        self.assertRaises(RuleError, self.handler.get_rule_number, 'test')



import unittest

from openformats.strings import OpenString

from openformats.tests.formats.common import CommonFormatTestMixin
from openformats.tests.utils.strings import (
    generate_random_string, strip_leading_spaces
)

from openformats.formats.android import AndroidHandler


class AndroidTestCase(CommonFormatTestMixin, unittest.TestCase):
    HANDLER_CLASS = AndroidHandler
    TESTFILE_BASE = "openformats/tests/formats/android/files"

    def setUp(self):
        super(AndroidTestCase, self).setUp()
        self.handler = AndroidHandler()

    def test_string(self):
        random_key = generate_random_string()
        random_string = generate_random_string()
        random_openstring = OpenString(random_key,
                                       random_string, order=0)
        random_hash = random_openstring.template_replacement

        source_python_template = u'''
            <resources>
                <string name="{key}">{string}</string>
            </resources>
        '''
        source = source_python_template.format(key=random_key,
                                               string=random_string)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [random_openstring])

        self.assertEquals(
            template,
            source_python_template.format(key=random_key, string=random_hash)
        )
        self.assertEquals(len(stringset), 1)
        self.assertEquals(stringset[0].__dict__, random_openstring.__dict__)
        self.assertEquals(compiled, source)

    def test_string_array(self):
        random_name = generate_random_string()
        random_key = '{}[0]'.format(random_name)
        random_string = generate_random_string()
        random_openstring = OpenString(random_key, random_string, order=0)
        random_hash = random_openstring.template_replacement
        source_python_template = strip_leading_spaces(u'''
            <resources>
                <string-array name="{key}">
                    <item>{string}</item>
                </string-array>
            </resources>
        ''')
        source = source_python_template.format(key=random_name,
                                               string=random_string)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [random_openstring])

        self.assertEquals(
            template,
            source_python_template.format(key=random_name, string=random_hash)
        )
        self.assertEquals(len(stringset), 1)
        self.assertEquals(stringset[0].__dict__,
                          random_openstring.__dict__)
        self.assertEquals(compiled, source)

    def test_plurals(self):
        random_key = generate_random_string()
        random_singular = generate_random_string()
        random_plural = generate_random_string()
        random_openstring = OpenString(random_key,
                                       {1: random_singular, 5: random_plural},
                                       order=0)
        random_hash = random_openstring.template_replacement

        source = strip_leading_spaces(u"""
            <resources>
                <plurals name="{key}">
                    <item quantity="one">{singular}</item>
                    <item quantity="other">{plural}</item>
                </plurals>
            </resources>
        """.format(key=random_key, singular=random_singular,
                   plural=random_plural))

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [random_openstring])

        self.assertEquals(
            template,
            strip_leading_spaces(u'''
                <resources>
                    <plurals name="{key}">
                        {hash_}
                    </plurals>
                </resources>
            '''.format(key=random_key, hash_=random_hash))
        )
        self.assertEquals(len(stringset), 1)
        self.assertEquals(stringset[0].__dict__, random_openstring.__dict__)
        self.assertEquals(compiled, source)

    def test_no_translatable(self):
        random_key = generate_random_string()
        random_string = generate_random_string()
        source = strip_leading_spaces(u'''
            <resources>
                <string name="{key}" translatable="false">{string}</string>
            </resources>
        '''.format(key=random_key, string=random_string))

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [])

        self.assertEquals(stringset, [])
        self.assertEquals(template, source)
        self.assertEquals(compiled, source)

    def test_order_is_kept(self):
        source = '''
            <resources>
                <string name="a">a</string>
                <string-array name="b">
                    <item>b</item>
                    <item>c</item>
                </string-array>
                <plurals name="c">
                    <item quantity="one">d</item>
                    <item quantity="other">e</item>
                </plurals>
            </resources>
        '''
        template, stringset = self.handler.parse(source)
        self.assertEquals([(string.key, string.order) for string in stringset],
                          [('a', 0), ('b[0]', 1), ('b[1]', 2), ('c', 3)])

    def test_string_inner_tags_are_collected(self):
        source = '''
            <resources>
                <string name="a">hello <b>world</b></string>
            </resources>
        '''
        template, stringset = self.handler.parse(source)
        string = stringset[0]
        self.assertEquals(string.key, 'a')
        self.assertEquals(string.string, "hello <b>world</b>")

    def test_string_content_is_valid_xml(self):
        source = '''
            <resources>
                <string name="a">hello <b>world</b></string>
            </resources>
        '''
        template, stringset = self.handler.parse(source)
        string = stringset[0]
        self.assertEquals(string.key, 'a')
        self.assertEquals(string.string, "hello <b>world</b>")

    def test_string_content_is_not_valid_xml(self):
        self._test_parse_error(
            '''
                <resources>
                    <string name="a">hello <b>world</c></string>
                </resources>
            ''',
            "Closing tag 'c' does not match opening tag 'b' on line 3"
        )

    def test_empty_string_ignored(self):
        random_key = generate_random_string()
        source = strip_leading_spaces(u'''
            <resources>
                <string name="{key}"></string>
            </resources>
        '''.format(key=random_key))

        template, stringset = self.handler.parse(source)

        self.assertEquals(stringset, [])
        self.assertEquals(template, source)

    def test_empty_string_array_item_ignored(self):
        random_key = generate_random_string()
        source = strip_leading_spaces(u'''
            <resources>
                <string-array name="{key}">
                    <item></item>
                </string-array>
            </resources>
        '''.format(key=random_key))

        template, stringset = self.handler.parse(source)

        self.assertEquals(stringset, [])
        self.assertEquals(template, source)

    def test_empty_plural_raises_error(self):
        self._test_parse_error(
            '<resources><plurals name="a"></plurals></resources>',
            u"Empty <plurals> tag on line 1"
        )

    def test_empty_plural_item_raises_error(self):
        self._test_parse_error(
            '''
                <resources>
                    <plurals name="a">
                        <item quantity="one"></item>
                        <item quantity="two"></item>
                        <item quantity="other">hello</item>
                    </plurals>
                </resources>
            ''',
            (
                'Missing string(s) in <item> tag(s) in the <plural> tag '
                'on line 3'
            )
        )

    def test_all_plural_items_empty_get_skipped(self):
        source = u'''
            <resources>
                <plurals name="a">
                    <item quantity="one"></item>
                    <item quantity="other"></item>
                </plurals>
            </resources>
        '''
        template, stringset = self.handler.parse(source)

        self.assertEquals(template, source)
        self.assertEquals(stringset, [])

    def test_missing_translated_strings_removed(self):
        random_key = generate_random_string()
        random_string = generate_random_string()
        source = strip_leading_spaces(u'''
            <resources>
                <string name="{key}">{string}</string>
            </resources>
        '''.format(key=random_key, string=random_string))

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [])
        self.assertEquals(compiled, strip_leading_spaces(u'''
            <resources>

            </resources>
        '''))

    def test_missing_translated_string_array_items_removed(self):
        random_key = generate_random_string()
        random_string1 = generate_random_string()
        random_string2 = generate_random_string()
        source = strip_leading_spaces(u'''
            <resources>
                <string-array name="{key}">
                    <item>{string1}</item>
                    <item>{string2}</item>
                </string-array>
            </resources>
        '''.format(key=random_key, string1=random_string1,
                   string2=random_string2))

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [stringset[0]])
        self.assertEquals(compiled, strip_leading_spaces(u'''
            <resources>
                <string-array name="{key}">
                    <item>{string1}</item>

                </string-array>
            </resources>
        '''.format(key=random_key, string1=random_string1)))

    def test_missing_translated_plurals_removed(self):
        random_key = generate_random_string()
        random_singular = generate_random_string()
        random_plural = generate_random_string()
        source = strip_leading_spaces(u'''
            <resources>
                <plurals name="{key}">
                    <item quantity="one">{singular}</item>
                    <item quantity="other">{plural}</item>
                </plurals>
            </resources>
        '''.format(key=random_key, singular=random_singular,
                   plural=random_plural))

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [])
        self.assertEquals(compiled, strip_leading_spaces(u'''
            <resources>

            </resources>
        '''.format(key=random_key, singular=random_singular,
                   plural=random_plural)))

    def test_missing_translated_string_arrays_removed(self):
        random_key = generate_random_string()
        random_string = generate_random_string()
        source = strip_leading_spaces(u'''
            <resources>
                <string-array name="{key}">
                    <item>{string}</item>
                </string-array>
            </resources>
        '''.format(key=random_key, string=random_string))

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [])
        self.assertEquals(compiled, strip_leading_spaces(u'''
            <resources>

            </resources>
        '''))

    def test_compile_plurals_not_indented(self):
        random_key = generate_random_string()
        random_singular = generate_random_string()
        random_plural = generate_random_string()
        random_openstring = OpenString(random_key,
                                       {1: random_singular, 5: random_plural},
                                       order=0)
        source = (u'<resources><plurals name="{key}"><item quantity="one">'
                  '{singular}</item><item quantity="other">{plural}</item>'
                  '</plurals></resources>').format(key=random_key,
                                                   singular=random_singular,
                                                   plural=random_plural)

        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [random_openstring])

        self.assertEquals(stringset[0].__dict__, random_openstring.__dict__)
        self.assertEquals(compiled, source)

    def test_missing_quantity_raises_error(self):
        self._test_parse_error(
            '''
                <resources>
                    <plurals name="a">
                        <item>a</item>
                        <item quantity="other">b</item>
                    </plurals>
                </resources>
            ''',
            u"Missing the `quantity` attribute on line 4"
        )

    def test_unknown_quantity(self):
        self._test_parse_error(
            '''
                <resources>
                    <plurals name="a">
                        <item quantity="one">a</item>
                        <item quantity="three-hundred-and-fifty-four">b</item>
                    </plurals>
                </resources>
            ''',
            u"The `quantity` attribute on line 5 contains an invalid plural: "
            u"`three-hundred-and-fifty-four`"
        )

    def test_missing_other_quantity(self):
        self._test_parse_error(
            '''
                <resources>
                    <plurals name="a">
                        <item quantity="one">a</item>
                        <item quantity="two">b</item>
                    </plurals>
                </resources>
            ''',
            u"Quantity 'other' is missing from <plurals> tag on line 3"
        )

    def test_wrong_tag_type_in_plurals_tag(self):
        self._test_parse_error(
            '''
                <resources>
                    <plurals name="a">
                        <item quantity="other">a</item>
                        <stuff quantity="one">b</stuff>
                    </plurals>
                </resources>
            ''',
            (
                u"Wrong tag type found on line 5. Was "
                u"expecting <item> but found <stuff>"
            )
        )

    def test_tail_characters_where_they_should_not_be(self):
        self._test_parse_error(
            '''
                <resources>
                    <plurals name="a">
                        <item quantity="other">a</item>My tail chars
                        <item quantity="one">b</item>
                    </plurals>
                </resources>
            ''',
            u"Found trailing characters after 'item' tag on line 4"
        )

    def test_duplicate_names(self):
        self._test_parse_error(
            '''
                <resources>
                    <string name="a">hello</string>
                    <string name="a">world</string>
                </resources>
            ''',
            u"Duplicate `name` (a) attribute found on line 4. Specify a "
            "`product` to differentiate"
        )

    def test_duplicate_names_and_products(self):
        self._test_parse_error(
            '''
                <resources>
                    <string name="a" product="b">hello</string>
                    <string name="a" product="b">>world</string>
                </resources>
            ''',
            u"Duplicate `name` (a) and `product` (b) attributes found on line "
            u"4"
        )

    def test_name_escaping_helps_with_duplication_cornercases(self):
        source = '''
            <resources>
                <string name="a[0]">hello</string>
                <string-array name="a">
                    <!-- Normally this string would get the name: 'a[0]' which
                         would be identical to the previous one -->
                    <item>goodbye</item>
                </string-array>
            </resources>
        '''
        template, stringset = self.handler.parse(source)
        self.assertEquals(stringset[0].key, "a\\[0]")
        self.assertEquals(stringset[1].key, "a[0]")

        source = r'''
            <resources>
                <string name="a[0]">hello</string>
                <!-- If we were only escaping the '[' char, this would still
                     confuse stuff since the item below would get the name:
                     `a\[0]` -->
                <string-array name="a\">
                    <item>goodbye</item>
                </string-array>
            </resources>
        '''
        template, stringset = self.handler.parse(source)
        self.assertEquals(stringset[0].key, "a\\[0]")
        self.assertEquals(stringset[1].key, "a\\\\[0]")

    def test_missing_name(self):
        self._test_parse_error('<resources><string>hello</string></resources>',
                               u"Missing the `name` attribute on line 1")

    def test_compile_removes_missing_strings(self):
        source = strip_leading_spaces('''
            <resources>
                <string name="a">hello</string>
                <string name="b">goodbye</string>
            </resources>
        ''')
        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, stringset[:1])
        self.assertEquals(
            compiled,
            strip_leading_spaces('''
                <resources>
                    <string name="a">hello</string>

                </resources>
            ''')
        )

    def test_compile_removes_missing_string_array_items(self):
        source = strip_leading_spaces('''
            <resources>
                <string-array name="a">
                    <item>hello</item>
                    <item>world</item>
                </string-array>
            </resources>
        ''')
        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, stringset[:1])
        self.assertEquals(
            compiled,
            strip_leading_spaces('''
                <resources>
                    <string-array name="a">
                        <item>hello</item>

                    </string-array>
                </resources>
            ''')
        )

    def test_compile_removes_missing_string_arrays(self):
        source = strip_leading_spaces('''
            <resources>
                <string-array name="a">
                    <item>hello</item>
                    <item>world</item>
                </string-array>
            </resources>
        ''')
        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [])
        self.assertEquals(
            compiled,
            strip_leading_spaces('''
                <resources>

                </resources>
            ''')
        )

    def test_compile_doesnt_remove_already_empty_string_array(self):
        source = strip_leading_spaces('''
            <resources>
                <string-array name="a"></string-array>
                <string-array name="b">
                    <item>hello</item>
                    <item>world</item>
                </string-array>
            </resources>
        ''')
        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [])
        self.assertEquals(
            compiled,
            strip_leading_spaces('''
                <resources>
                    <string-array name="a"></string-array>

                </resources>
            ''')
        )

    def test_compile_removes_missing_plurals(self):
        source = strip_leading_spaces('''
            <resources>
                <plurals name="a">
                    <item quantity="one">hello</item>
                    <item quantity="other">world</item>
                </plurals>
            </resources>
        ''')
        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [])
        self.assertEquals(
            compiled,
            strip_leading_spaces('''
                <resources>

                </resources>
            ''')
        )

    def test_compile_doesnt_remove_already_empty_plurals(self):
        source = strip_leading_spaces('''
            <resources>
                <plurals name="a">
                    <item quantity="one">hello</item>
                    <item quantity="other">world</item>
                </plurals>
                <plurals name="b">
                    <item quantity="one"></item>
                    <item quantity="other"></item>
                </plurals>
            </resources>
        ''')
        template, stringset = self.handler.parse(source)
        compiled = self.handler.compile(template, [])
        self.assertEquals(
            compiled,
            strip_leading_spaces('''
                <resources>

                    <plurals name="b">
                        <item quantity="one"></item>
                        <item quantity="other"></item>
                    </plurals>
                </resources>
            ''')
        )

    def test_parser_doesnt_like_text_where_it_shouldnt_be(self):
        self._test_parse_error(
            u'<resources>hello<string name="a">world</string></resources>',
            u"Found leading characters inside 'resources' tag on line 1"
        )

    def test_strings_from_plurals_are_always_pluralized(self):
        _, stringset = self.handler.parse(
            u'''
                <resources>
                    <plurals name="a">
                        <item quantity="other">hello</item>
                    </plurals>
                </resources>
            '''
        )
        self.assertTrue(stringset[0].pluralized)

    def test_escape(self):
        cases = (('double " quote', 'double \\" quote'),
                 ("single ' quote", "single \\' quote"),
                 ("back \\ slash", "back \\\\ slash"))
        for rich, raw in cases:
            self.assertEquals(AndroidHandler.escape(rich), raw)

    def test_unescape(self):
        cases = (('double " quote', 'double \\" quote'),
                 ("single ' quote", "single \\' quote"),
                 ("back \\ slash", "back \\\\ slash"),
                 ('inside double quotes', '"inside double quotes"'),
                 ("single ' quote", '"single \' quote"'),
                 ("back \\ slash", '"back \\ slash"'))
        for rich, raw in cases:
            self.assertEquals(AndroidHandler.unescape(raw), rich)



"""
Various methods useful for tests and similar operations.
"""

import csv
import os


root = os.path.dirname(__file__)
DICT_FNAME = os.path.join(root, 'dictionary.csv')


class FunkyDictionary(object):
    def __init__(self):
        self.phrase_list = []
        self.phrase_dict = {}
        with open(DICT_FNAME, 'rU') as dict_file:
            dict_reader = csv.DictReader(dict_file)
            for phrase in dict_reader:
                unicode_phrase = {key: value.decode("utf-8")
                                  for key, value in phrase.items()}
                self.phrase_list.append(unicode_phrase)
                # We can assume 'en' is going to be used as a source language
                # often, so it makes sense to be able to do quick lookups
                self.phrase_dict[unicode_phrase['en']] = unicode_phrase

    def translate(self, phrase, to_lang, from_lang="en", debug=False):
        if from_lang == "en":
            try:
                if debug:
                    print('Lookup for "{}" successful'.format(phrase[:20]))
                return self.phrase_dict[phrase][to_lang]
            except KeyError:
                pass
        else:
            for dict_phrase in self.phrase_list:
                if phrase == dict_phrase[from_lang]:
                    return dict_phrase[to_lang]
        # Phrase not found in funky dict
        if debug:
            print('Lookup for "{}" unsuccessful.'.format(phrase[:20]))
        return u"{}:{}".format(to_lang, phrase)


funky_dictionary = FunkyDictionary()


def translate_stringset(stringset, from_lang="en", to_lang="el", debug=False):
    for s in stringset:
        for rule, pluralform in s._strings.items():
            s._strings[rule] = funky_dictionary.translate(
                pluralform, to_lang, from_lang, debug
            )
    return stringset



from __future__ import absolute_import

import json
import re


class DumbJson(object):
    """ A utility to help iterate over a JSON string. The main focuses are:

        1. Return the exact contents of each encountered string, don't unescape
           double quotes ('"')
        2. Also return the positions of things encountered

        To initialize, simply pass a JSON string:

            >>> dumb_json = DumbJson('{"hello": "world"}')

        If you want, you can pass an extra argument to identify an embedded
        JSON object within the outer one. For example, if you have this string.

            >>> source = '["first string", {"second": "dict"}, "third string"]'

        You can:

            >>> start = source.index('{')  # 17
            >>> dumb_json = DumbJson(source, start)

        In this case, when you iterate over this, it will only yield the inner
        dictionary (`{"second": "string"}`). The item positions yielded while
        iterating will be in respect to the outer string, so:

            >>> assert list(dumb_json) == [('second', 19, 'dict', 29)]

        If the DumbJson object is a dictionary, then iterating it will yield
        4-tuples with `(key, key_position, value, value_position)`. If it's a
        list it will yield 2-tuples with `(item, item_position)`. Eg:

            >>> assert list(DumbJson('{"a": "b"}')) == [('a', 2, 'b', 7)]
            >>> assert list(DumbJson('["a", "b"]')) == [('a', 2), ('b', 7)]

        Encountering an embedded JSON structure while iterating will yield a
        DumbJson object:

            >>> embedded, _ = list(DumbJson('[["a"]]'))[0]
            >>> assert isinstance(embedded, DumbJson)
            >>> assert list(embedded) == [("a", 3)]

            # Note that the position (3) is in respect to the root JSON string

        When the items or values are not strings but objects allowed by JSON,
        like numbers, booleans or null, they will be yielded normally:

            >>> assert list(DumbJson('{"a": null}')) == [("a", 2, None, 6)]
            >>> assert list(DumbJson('[null]')) == [(None, 2)]
    """

    def __init__(self, source, start=0):
        self.source = source
        self._end = None
        starting_symbol, self.start = self._find_next('{[', start)
        if starting_symbol == '{':
            self.type = dict
        elif starting_symbol == '[':
            self.type = list
        else:
            raise ValueError("Input is not a JSON container")

    def __iter__(self):
        if self.type == dict:
            return self._iter_dict()
        elif self.type == list:
            return self._iter_list()

    def _iter_dict(self):
        # The '_p' suffix means 'position'

        start = self.start + 1

        # Maybe it's an empty dict
        end, end_p = self._find_next('"}', start)
        if end == "}":
            self.end = end_p
            return

        while True:
            # Lets find our key
            _, start_key_quote_p = self._find_next('"', start)
            key_p = start_key_quote_p + 1
            _, end_key_quote_p = self._find_next('"', key_p,
                                                 require_whitespace=False)
            key = self.source[key_p:end_key_quote_p]
            _, colon_p = self._find_next(':', end_key_quote_p + 1)
            value_start_string, value_start_computed, value_start_p =\
                self._process_value(colon_p + 1)

            # Our job in each case is to yield something and set 'next_p' to
            # where we should search for our next item
            if value_start_string == '"':
                # We found a string!
                value_p = value_start_p + 1
                _, value_end_quote_p = self._find_next(
                    '"', value_p, require_whitespace=False
                )
                value = self.source[value_p:value_end_quote_p]
                yield key, key_p, value, value_p
                next_p = value_end_quote_p + 1
            elif value_start_string in ('{', '['):
                # We found an embedded, lets return an instance of ourself
                embedded = DumbJson(self.source, value_start_p)
                yield key, key_p, embedded, value_start_p
                next_p = embedded.end + 1
            elif (value_start_computed is not None or
                    value_start_string == "null"):
                # We found something else allowed by JSON
                yield key, key_p, value_start_computed, value_start_p
                next_p = value_start_p + len(value_start_string)
            else:
                # Something went wrong
                raise ValueError("No JSON value could be decoded")

            next_symbol, next_symbol_p = self._find_next(',}', next_p)
            if next_symbol == ',':
                start = next_symbol_p + 1
            elif next_symbol == '}':
                self.end = next_symbol_p
                break

    def _iter_list(self):
        # The '_p' suffix means 'position'

        start = self.start + 1

        # Maybe it's an empty list
        match = re.search(r'^\s*.', self.source[start:])
        if match:
            if match.group()[-1] == "]":
                self.end = start + match.end() - 1
                return

        while True:
            # Lets find our items
            item_start_string, item_start_computed, item_start_p =\
                self._process_value(start)

            # Our job in each case is to yield something and set 'next_p' to
            # where we should search for our next item
            if item_start_string == '"':
                # We found a string!
                item_p = item_start_p + 1
                _, end_item_quote_p = self._find_next('"', item_p,
                                                      require_whitespace=False)
                item = self.source[item_p:end_item_quote_p]
                yield item, item_p
                next_p = end_item_quote_p + 1
            elif item_start_string in ('{', '['):
                # We found an embedded, lets return an instance of ourself
                embedded = DumbJson(self.source, item_start_p)
                yield embedded, item_start_p
                next_p = embedded.end + 1
            elif (item_start_computed is not None or
                    item_start_string == "null"):
                # We found something else allowed by JSON
                yield item_start_computed, item_start_p
                next_p = item_start_p + len(item_start_string)
            else:
                # Something went wrong
                raise ValueError("No JSON value could be decoded")

            next_symbol, next_symbol_p = self._find_next(',]', next_p)
            if next_symbol == ',':
                start = next_symbol_p + 1
            elif next_symbol == ']':
                self.end = next_symbol_p
                break

    def _find_next(self, symbols, start=0, require_whitespace=True):
        symbols = {s for s in symbols}
        for ptr in xrange(start, len(self.source)):
            candidate = self.source[ptr]
            if candidate in symbols:
                if (candidate == '"' and ptr > 0 and
                        self.source[ptr - 1] == '\\'):
                    continue
                return candidate, ptr
            if require_whitespace and not candidate.isspace():
                newline_count = self.source.count('\n', 0, ptr)
                raise ValueError(
                    u"Was expecting whitespace or one of `{symbols}` on line "
                    u"{line_no}, found `{candidate}` instead".format(
                        symbols=''.join(symbols),
                        line_no=newline_count + 1,
                        candidate=candidate,
                    )
                )
        return None, None

    def _process_value(self, start):
        """ A variation of _find_next. If the next non-empty character after
            `start` is in ('"', '{', '['), this will behave exactly like
            `_find_next('"{[', start)`. If the next non empty sequence after
            `start` is a number, bolean or 'null', it will return appropriate
            values.

            Returns 3 values:
            - value_start_string: in case of ('"', '{', '['), it is the same as
              the first return value of _find_next. Otherwise, it's the string
              representation of whatever the value is (the actual string "true"
              or "3.14159")
            - value_start_computed: in case of ('"', '{', '['), it is
              irrelevant (None), otherwise it's the computed value
            - value_start_p: where the value, whatever it is, is encountered
        """

        # Lets construct a regular expression to find the first non-empty char
        value_pat = r'{dict_list_string}|{true_false_null}|{e_notation}|'\
            r'{_float}|{integer}'.format(
                dict_list_string=r'[{\["]',
                true_false_null=r'true|false|null',
                e_notation=r'-?\d+e-?\d+',
                _float=r'-?\d+\.\d+',
                integer=r'-?\d+',
            )
        first_non_empty = r'^(?P<spaces>\s*)(?P<value>{})'.format(value_pat)
        match = re.search(first_non_empty, self.source[start:])
        # We probably found a match, otherwise this is not JSON
        if match:
            spaces, value = match.groups()
            value_start = start + len(spaces)
            if value in ('{', '[', '"'):
                return value, None, value_start
            else:
                # We either have true/false/null or a number of sorts
                return value, json.loads(value), value_start
        else:
            raise ValueError("No JSON value could be decoded")

    @property
    def end(self):
        if self._end is None:
            # In order for 'end' to be calculated, 'self' must be iterated over
            # first. Normally this should happen on its own when we're
            # searching in a DFS manner, otherwise, we have to force it.
            for _ in self:
                pass
        return self._end

    @end.setter
    def end(self, value):
        self._end = value



from __future__ import absolute_import

from ..handlers import Handler
from openformats.strings import OpenString
from ..utils.compilers import OrderedCompilerMixin


class PlaintextHandler(OrderedCompilerMixin, Handler):
    name = "Plaintext"
    extension = "txt"

    def parse(self, content):
        stringset = []
        # find out whether we're using UNIX or DOS newlines
        try:
            position = content.index('\n')
        except ValueError:
            # No newlines present
            newline_sequence = ""
            lines = (content, )
        else:
            if position == 0 or content[position - 1] != '\r':
                newline_sequence = "\n"
            else:
                newline_sequence = "\r\n"
            lines = content.split(newline_sequence)

        template = ""
        order = 0
        for line in lines:
            stripped_line = line.strip()
            if stripped_line:
                string = OpenString(str(order), stripped_line, order=order)
                order += 1
                stringset.append(string)

                template_line = line.replace(stripped_line,
                                             string.template_replacement)
            template += newline_sequence
            if stripped_line:
                template += template_line
            else:
                template += line

        # remove newline_sequence added to the start of the template
        template = template[len(newline_sequence):]
        return template, stringset



from openformats.exceptions import RuleError


class Handler(object):
    """
    This class defines the interface you need to implement in order to create a
    handler. Both the `parse` and `compile` methods must be implemented.
    """

    name = None
    extension = None

    _RULES_ATOI = {
        'zero': 0,
        'one': 1,
        'two': 2,
        'few': 3,
        'many': 4,
        'other': 5
    }

    EXTRACTS_RAW = True

    _RULES_ITOA = {value: key for key, value in _RULES_ATOI.iteritems()}

    _RULE_ERROR_MSG = (
        '{attempted} is not a valid rule value. Valid choices are {valid}'
    )

    @classmethod
    def get_rule_number(cls, string_value):
        try:
            return cls._RULES_ATOI[string_value]
        except KeyError:
            msg = cls._RULE_ERROR_MSG.format(
                attempted=string_value, valid=cls._RULES_ATOI.keys()
            )
            raise RuleError(msg)

    @classmethod
    def get_rule_string(cls, number_value):
        try:
            return cls._RULES_ITOA[number_value]
        except KeyError:
            msg = cls._RULE_ERROR_MSG.format(
                attempted=number_value, valid=cls._RULES_ITOA.keys()
            )
            raise RuleError(msg)

    def parse(self, content):
        """
        Parses the content, extracts translatable strings into a stringset,
        replaces them with hashes and returns a tuple of the template with the
        stringset

        Typically this is done in the following way:

        * Use a library or your own code to segment (deserialize) the content
          into translatable entities.
        * Choose a key to uniquely identify the entity.
        * Create an ``OpenString`` object representing the entity.
        * Create a hash to replace the original content with.
        * Create a stringset with the content.
        * Use library or own code to serialize stringset back into a template.
        """

        raise NotImplemented('Abstract method')  # pragma: no cover

    def compile(self, template, stringset):
        """
        Parses the template, finds the hashes, replaces them with strings from
        the stringset and returns the compiled file. If a hash in the template
        isn't found in the stringset, it's a good practice to remove the whole
        string section surrounding it

        Typically this is done in the following way:

        * Use a library or own code to segment (deserialize) the template into
          translatable entities, as if assuming that the hashes are the
          translatable entities.
        * Make sure the hash matches the first string in the stringset.
        * Replace the hash with the string.
        * Use library or own code to serialize stringset back into a compiled
          file.

        You can safely assume that the stringset will have strings in the
        correct order for the above process and thus you will probably be able
        to perform the whole compilation in a single pass.
        """

        raise NotImplemented('Abstract method')  # pragma: no cover



from __future__ import absolute_import

import itertools

from ..handlers import Handler
from ..strings import OpenString
from ..utils.xml import NewDumbXml, DumbXmlSyntaxError
from ..exceptions import RuleError
from ..exceptions import ParseError
from ..transcribers import Transcriber


class AndroidHandler(Handler):
    """A handler class that parses and compiles String Resources for ANDROID
    applications. The String Resources file is in XML format.

    String Resources file documentation can be found here:
    http://developer.android.com/guide/topics/resources/string-resource.html
    """

    name = "ANDROID"
    extension = "xml"

    EXTRACTS_RAW = True

    # Where to start parsing the file
    PARSE_START = "<resources"

    # Relevant tags
    STRING = "string"
    STRING_PLURAL = "plurals"
    STRING_ARRAY = "string-array"

    # Relevant children
    STRING_ITEM = "item"

    # Attributes that if the child contains it should be skipped
    SKIP_ATTRIBUTES = {
        'translatable': 'false'
    }

    # Compile plural template
    PLURAL_TEMPLATE = u'<item quantity="{rule}">{string}</item>'

    """ Parse Methods """

    def parse(self, content):
        self.transcriber = Transcriber(content)
        self.current_comment = u""
        self.order_counter = itertools.count()

        source = self.transcriber.source
        # Skip xml info declaration
        resources_tag_position = content.index(self.PARSE_START)
        try:
            parsed = NewDumbXml(source, resources_tag_position)
            self._validate_no_text_characters(parsed)
            self._validate_no_tail_characters(parsed)
            children_itterator = parsed.find_children(
                self.STRING,
                self.STRING_ARRAY,
                self.STRING_PLURAL,
                NewDumbXml.COMMENT
            )
            stringset = []
            self.existing_hashes = set()
            for child in children_itterator:
                strings = self._handle_child(child)
                if strings is not None:
                    stringset.extend(strings)
                    self.current_comment = u""
        except DumbXmlSyntaxError, e:
            raise ParseError(unicode(e))

        self.transcriber.copy_until(len(source))
        template = self.transcriber.get_destination()

        return template, stringset

    def _handle_child(self, child):
        """Do basic checks on the child and assigns the appropriate method to
            handle it based on the child's tag.

        :returns: An list of OpenString objects if any were created else None.
        """
        self._validate_no_tail_characters(child)
        if not self._should_ignore(child):
            if child.tag == NewDumbXml.COMMENT:
                self._handle_comment(child)
            else:
                if child.tag == self.STRING:
                    return self._handle_string(child)
                elif child.tag == self.STRING_ARRAY:
                    self._validate_no_text_characters(child)
                    return self._handle_string_array(child)
                elif child.tag == self.STRING_PLURAL:
                    self._validate_no_text_characters(child)
                    return self._handle_string_plural(child)
        return None

    def _handle_string(self, child):
        """Handles child element that has the `string` tag.

        If it contains a string it will create an OpenString object.

        :returns: An list of containing the OpenString object
                    if one was created else it returns None.
        """
        name, product = self._get_child_attributes(child)
        string = self._create_string(
            name,
            child.content,
            self.current_comment,
            product,
            child
        )
        if string is not None:
            # <string>My Text</string>
            #         ^
            self.transcriber.copy_until(child.text_position)
            self.transcriber.add(string.template_replacement)
            # <string>My Text</string>
            #                ^
            self.transcriber.skip(len(child.content))
            return [string]
        return None

    def _handle_string_plural(self, child):
        """Handles child element that has the `plurals` tag.

        It will find children with the `item` tag and create an OpenString
        object out of them.

        :raises: Parse error if the `quantity` attribute is missing from any
                    of the child's children
        :returns: An list containing the OpenString object if one was created
                    else None.
        """

        string_rules_text = {}
        item_itterator = child.find_children()
        # Itterate through the children with the item tag.
        for item_tag in item_itterator:
            rule_number = self._validate_plural_item(item_tag)
            string_rules_text[rule_number] = item_tag.content

        name, product = self._get_child_attributes(child)
        string = self._create_string(
            name,
            string_rules_text,
            self.current_comment,
            product,
            child,
            # <plurals> tags always define plurals, even if the language has
            # one plural form and thus there's only one <item>
            pluralized=True,
        )
        if string is not None:
            # <plurals name="foo">   <item>Hello ...
            #                        ^
            first_plural_position = child.text_position + len(child.text)
            self.transcriber.copy_until(first_plural_position)
            self.transcriber.add(string.template_replacement)
            # ...</item>   </plurals>...
            #           ^
            self.transcriber.skip_until(item_tag.tail_position)
            # FYI: item_tag is the last iterated item from the loop before.
            return [string]
        return None

    def _handle_string_array(self, child):
        """Handles child element that has the `string-array` tag.

        It will find children with the `item` tag and create an OpenString
        object out of each one of them.

        :returns: An list containing the OpenString objects if any were created
                    else None.
        """
        strings = []
        item_itterator = child.find_children(self.STRING_ITEM)
        name, product = self._get_child_attributes(child)
        # Itterate through the children with the item tag.
        for index, item_tag in enumerate(item_itterator):
            self._validate_no_tail_characters(item_tag)
            child_name = u"{}[{}]".format(name, index)
            string = self._create_string(
                child_name,
                item_tag.content,
                self.current_comment,
                product,
                child
            )

            if string is not None:
                # ... <item>Hello...
                #           ^
                self.transcriber.copy_until(item_tag.text_position)

                strings.append(string)
                self.transcriber.add(string.template_replacement)

                # ...ello world</item>...
                #              ^
                self.transcriber.skip(len(item_tag.content))
        if strings:
            return strings
        return None

    def _handle_comment(self, child):
        """Will assign the comment found as the current comment."""
        self.current_comment = child.content

    def _create_string(self, name, text, comment, product, child,
                       pluralized=False):
        """Creates a string and returns it. If empty string it returns None.

        :param text: The strings text.
        :param name: The name of the string.
        :param comment: The developer's comment the string might have.
        :param product: Extra context for the string.
        :param child: The child tag that the string is created from.
                        Used to find line numbers when errors occur.
        :returns: Returns an OpenString object if the text is not empty
                  else None.
        """
        if self._validate_not_empty(text, child):
            # Create OpenString
            string = OpenString(
                name,
                text,
                context=product,
                order=self.order_counter.next(),
                developer_comment=comment,
                pluralized=pluralized,
            )
            # If duplicate hash raise ParseError
            if string.string_hash in self.existing_hashes:
                format_dict = {'name': name}
                if not product:
                    msg = (
                        u"Duplicate `name` ({name}) attribute found on line "
                        u"{line_number}. Specify a `product` to differentiate"
                    )
                else:
                    format_dict['product'] = product
                    msg = (
                        u"Duplicate `name` ({name}) and `product` ({product}) "
                        u"attributes found on line {line_number}"
                    )
                self._raise_error(child, msg, context=format_dict)
            self.existing_hashes.add(string.string_hash)
            return string
        return None

    def _validate_plural_item(self, item_tag):
        """ Performs a number of checks on the plural item to see its validity.

        :param item_tag: The item to perform the checks on.
        :raises: ParseError if the item tag does not meet the requirments.
        :returns: The plural number of the validated item tag.
        """
        if item_tag.tag != self.STRING_ITEM:
            msg = (
                u"Wrong tag type found on line {line_number}. Was "
                u"expecting <item> but found <{wrong_tag}>"
            )
            self._raise_error(
                item_tag, msg, context={'wrong_tag': item_tag.tag}
            )

        self._validate_no_tail_characters(item_tag)

        rule = item_tag.attrib.get('quantity')
        if rule is None:
            # If quantity is missing, the plural is unknown
            msg = u"Missing the `quantity` attribute on line {line_number}"
            self._raise_error(item_tag, msg)
        try:
            rule_number = self.get_rule_number(rule)
        except RuleError:
            msg = (
                u"The `quantity` attribute on line {line_number} contains "
                u"an invalid plural: `{rule}`"
            )
            self._raise_error(item_tag, msg, context={'rule': rule})
        return rule_number

    def _validate_not_empty(self, text, child):
        """Validates that a string is not empty.

        :param text: The string to validate. Can be a basestring or a dict for
                        for pluralized strings.
        :param child: The tag that the string is created from.Used to show the
                        appropriate line number in case an error occurs.
        :raises: Raises a ParseError if not all plurals of a pluralized string
                 are complete.
        :returns: True if the string is not empty else False.
        """
        # If dict then it's pluralized
        if isinstance(text, dict):
            if len(text) == 0:
                msg = u"Empty <plurals> tag on line {line_number}"
                self._raise_error(child, msg)

            # Find the plurals that have empty string
            text_value_set = set(value.strip() for value in text.itervalues())
            if "" in text_value_set and len(text_value_set) != 1:
                # If not all plurals have empty strings raise ParseError
                msg = (
                    u'Missing string(s) in <item> tag(s) in the <plural> tag '
                    u'on line {line_number}'
                )
                self._raise_error(child, msg)
            elif "" in text_value_set:
                # All plurals are empty so skip `plurals` tag
                return False

            # Validate `other` rule is present
            if self._RULES_ATOI['other'] not in text.keys():
                msg = (
                    u"Quantity 'other' is missing from <plurals> tag "
                    u"on line {line_number}"
                )
                self._raise_error(child, msg)
        elif text.strip() == "":
            return False
        return True

    def _get_child_attributes(self, child):
        """Retrieves child's `name` and `product` attributes.

        :param child: The child to retrieve the attributes from.
        :returns: Returns a tuple (`name`, `product`)
        :raises: It raises a ParseError if no `name` attribute is present.
        """
        name = child.attrib.get('name')
        if name is None:
            msg = u'Missing the `name` attribute on line {line_number}'
            self._raise_error(child, msg)
        name = name.replace('\\', '\\\\').replace('[', '\\[')
        product = child.attrib.get('product', '')
        return name, product

    def _validate_no_tail_characters(self, child):
        if child.tail.strip() != "":
            # Check for tail characters
            self.transcriber.copy_until(child.tail_position)
            msg = (u"Found trailing characters after '{tag}' tag on line "
                   u"{line_number}")
            tag = child.tag
            if tag == NewDumbXml.COMMENT:
                tag = u"comment"
            self._raise_error(
                child,
                msg,
                context={'tag': tag,
                         'line_number': self.transcriber.line_number}
            )

    def _validate_no_text_characters(self, child):
        if child.text.strip() != "":
            # Check for text characters
            self.transcriber.copy_until(child.text_position)
            msg = (u"Found leading characters inside '{tag}' tag on line "
                   u"{line_number}")
            self._raise_error(
                child, msg,
                context={'tag': child.tag,
                         'line_number': self.transcriber.line_number}
            )

    def _raise_error(self, child, message, context=None):
        context = context or {}
        if 'line_number' not in context:
            self.transcriber.copy_until(child.position)
            context['line_number'] = self.transcriber.line_number
        raise ParseError(message.format(**context))

    """ Compile Methods """

    def compile(self, template, stringset):
        resources_tag_position = template.index(self.PARSE_START)

        self.transcriber = Transcriber(template[resources_tag_position:])
        source = self.transcriber.source

        parsed = NewDumbXml(source)
        children_itterator = parsed.find_children(
            self.STRING,
            self.STRING_ARRAY,
            self.STRING_PLURAL
        )

        self.stringset = iter(stringset)
        self.next_string = self._get_next_string()
        for child in children_itterator:
            self._compile_child(child)

        self.transcriber.copy_until(len(source))
        compiled = template[:resources_tag_position] +\
            self.transcriber.get_destination()

        return compiled

    def _compile_child(self, child):
        """Do basic checks on the child and assigns the appropriate method to
            handle it based on the child's tag.
        """
        if not self._should_ignore(child):
            if child.tag == self.STRING:
                self._compile_string(child)
            elif child.tag == self.STRING_ARRAY:
                self._compile_string_array(child)
            elif child.tag == self.STRING_PLURAL:
                self._compile_string_plural(child)

    def _compile_string(self, child):
        """Handles child element that has the `string` and `item` tag.

        It will compile the tag if matching string exists. Otherwise it will
        skip it.
        """
        if self._should_compile(child):
            self.transcriber.copy_until(child.text_position)
            self.transcriber.add(self.next_string.string)
            self.transcriber.skip_until(child.content_end)
            self.next_string = self._get_next_string()
        else:
            self._skip_tag(child)

    def _compile_string_array(self, child):
        """Handles child element that has the `string-array` tag.

        It will find children with the `item` tag that should be compiled and
        will compile them. If no matching string is found for a child it will
        remove it. If the `string-array` tag will be empty after compilation
        it will remove it as well.

        :NOTE: If the `string-array` was empty to begin with it will leave it
                as it is.
        """
        item_itterator = list(child.find_children(self.STRING_ITEM))

        # If placeholder (has no children) skip
        if len(item_itterator) == 0:
            return

        # Check if any string matches array items
        has_match = False
        for item_tag in item_itterator:
            if self._should_compile(item_tag):
                has_match = True
                break

        if has_match:
            # Compile found item nodes. Remove the rest.
            for item_tag in item_itterator:
                self._compile_string(item_tag)
        else:
            # Remove the `string-array` tag
            self._skip_tag(child)

    def _compile_string_plural(self, child):
        """Handles child element that has the `plurals` tag.

        It will check if pluralized string exists and add every plural as an
        `item` child. If no matching string is found it will remove the tag.

        :NOTE: If the `plurals` had empty `item` tags to begin with we leave
                it as it is.
        """
        # If placeholder (has empty children) skip
        if len(list(child.find_children(self.STRING_ITEM))):
            return

        if self._should_compile(child):
            self.transcriber.copy_until(child.text_position)

            splited_content = child.content.split(
                self.next_string.template_replacement
            )
            start = splited_content[0]
            end = splited_content[1]

            # If newline formating
            if start.startswith(end):
                start = start.replace(end, '', 1)
                self.transcriber.add(end)

            for rule, string in self.next_string.string.items():
                self.transcriber.add(
                    start +
                    self.PLURAL_TEMPLATE.format(
                        rule=self.get_rule_string(rule), string=string
                    )
                    + end
                )
            self.transcriber.skip_until(child.content_end)
            self.next_string = self._get_next_string()
        else:
            self._skip_tag(child)

    def _should_compile(self, child):
        """Checks if the current child should be compiled.

        :param child: The child to check if it should be compiled.
        :returns: True if the child should be compiled else False.
        """
        return (
            self.next_string is not None and
            self.next_string.template_replacement == child.content.strip()
        )

    def _skip_tag(self, child):
        """Skips a tag from the compilation.

        :param child: The tag to be skipped.
        """
        self.transcriber.copy_until(child.position)
        self.transcriber.skip_until(child.tail_position)

    def _get_next_string(self):
        """Gets the next string from stringset itterable.

        :returns: An openstring object or None if it has reached the end of
                    the itterable.
        """
        try:
            next_string = self.stringset.next()
        except StopIteration:
            next_string = None
        return next_string

    """ Util Methods """

    @staticmethod
    def _should_ignore(child):
        """Checks if the child contains any key:value pair from the
            SKIP_ATTRIBUTES dict.

        :returns: True if it contains any else false.
        """
        for key, value in AndroidHandler.SKIP_ATTRIBUTES.iteritems():
            filter_attr = child.attrib.get(key)
            if filter_attr is not None and filter_attr == value:
                return True
        return False

    # Escaping / Unescaping
    # According to:
    # http://developer.android.com/guide/topics/resources/string-resource.html#FormattingAndStyling  # noqa

    @staticmethod
    def escape(string):
        return string.replace('\\', '\\\\').replace('"', '\\"').\
            replace("'", "\\'")

    @staticmethod
    def unescape(string):
        if len(string) and string[0] == string[-1] == '"':
            return string[1:-1]
        else:
            return string.replace('\\"', '"').replace("\\'", "'").\
                replace('\\\\', '\\')



from hashlib import md5
from random import randint
from unittest import TestCase

from openformats.strings import OpenString

from openformats.tests.utils import generate_random_string


class OpenStringTestCase(TestCase):
    def test_singular_string_is_non_plural(self):
        random_string = generate_random_string()
        open_string = OpenString('test', random_string)

        self.assertEqual(open_string._strings, {5: random_string})
        self.assertFalse(open_string.pluralized)

    def test_singular_dict_is_non_plural(self):
        random_string = {5: generate_random_string()}
        open_string = OpenString('test', random_string)

        self.assertEqual(open_string._strings, random_string)
        self.assertFalse(open_string.pluralized)

    def test_singular_is_returned(self):
        random_string = generate_random_string()
        open_string = OpenString('test', random_string)

        self.assertEqual(open_string.string, random_string)

    def test_plurals_are_returned(self):
        random_strings = {
            i: generate_random_string()
            for i in xrange(randint(10, 20))
            }
        open_string = OpenString('test', random_strings)

        self.assertTrue(open_string.pluralized)
        self.assertEqual(open_string.string, random_strings)

    def test_multiples_are_plural(self):
        test_strings = {
            i: generate_random_string()
            for i in xrange(randint(5, 10))
            }
        open_string = OpenString('test', test_strings)

        self.assertEqual(open_string._strings, test_strings)
        self.assertTrue(open_string.pluralized)

    def test_defaults_are_assigned_to_self(self):
        random_property_name = generate_random_string()
        random_default = generate_random_string()

        backup_defaults = OpenString.DEFAULTS
        OpenString.DEFAULTS = {
            random_property_name: random_default
        }
        open_string = OpenString('test', 'test')
        self.assertEqual(
            getattr(open_string, random_property_name), random_default
        )
        OpenString.DEFAULTS = backup_defaults

    def test_pluralized_is_overridden_with_keyword(self):
        open_string = OpenString('test', 'test', pluralized=True)
        self.assertTrue(open_string.pluralized)

    def test_template_replacement_returns_correct_suffix(self):
        open_string = OpenString('test', 'test')

        # Test when pluralized False
        open_string.pluralized = False
        template_replacement = open_string.template_replacement
        self.assertTrue(template_replacement.endswith('tr'))

        # Test when pluralized True
        open_string.pluralized = True
        template_replacement = open_string.template_replacement
        self.assertTrue(template_replacement.endswith('pl'))

    def test_context_is_hashed_when_present(self):
        random_context = generate_random_string()
        random_key = generate_random_string()
        random_hash = md5(
            ':'.join([random_key, random_context]).encode('utf-8')
        ).hexdigest()

        open_string = OpenString(random_key, 'test')
        open_string.context = random_context

        replacement = open_string.template_replacement
        hash_string = replacement.split('_')[0]

        self.assertEqual(hash_string, random_hash)

    def test_hash_is_calculated_from_components(self):
        random_key = generate_random_string()
        random_context = generate_random_string()
        random_rule = randint(1, 5)
        random_hash = hash((random_key, random_context, random_rule))

        open_string = OpenString(random_key, '')
        open_string.context = random_context
        open_string.rule = random_rule

        self.assertEqual(hash(open_string), random_hash)



#!/usr/bin/env python

"""
Create template files from source files using the respective handlers.

Example:
    $ ./bin/create_files.py openformats/tests/srt/files/1_en.srt
"""

from __future__ import absolute_import

import argparse
import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

from openformats.formats import plaintext, srt, android, json
from openformats.tests.utils import translate_stringset

args = argparse.ArgumentParser


def get_handler(ext):
    """Return the right format handler based on the file extension."""
    return {
        'txt': plaintext.PlaintextHandler(),
        'srt': srt.SrtHandler(),
        'xml': android.AndroidHandler(),
        'json': json.JsonHandler(),
    }[ext]


def run():
    # Choose correct handler based on the file extension
    file_extension = os.path.splitext(args.inputfile)[1][1:]
    handler = get_handler(file_extension)

    with open(args.inputfile, mode='rU') as f:
        source_contents = f.read()

    # Save template test file, eg. 1_tpl.foo
    template, stringset = handler.parse(source_contents)
    tpl_fname = args.inputfile.replace("_en", "_tpl")
    with open(tpl_fname, 'w+') as tpl_file:
        if args.debug:
            print("Writing %s" % tpl_fname)
        if args.execute:
            tpl_file.write(template)
        tpl_file.close()

    translated_stringset = translate_stringset(stringset, debug=True)

    # Save translated file
    compiled = handler.compile(template, translated_stringset)
    fname = args.inputfile.replace("_en", "_el")
    with open(fname, 'w+') as f:
        if args.debug:
            print("Writing %s" % fname)
        f.write(compiled)
        f.close()


def main(argv):
    parser = argparse.ArgumentParser(
        add_help=True,
        description='Generate right test files from an English source file.'
    )
    parser.add_argument('inputfile',
                        help="Source file to convert")
    parser.add_argument('-d', '--debug', action='store_true', default=True,
                        help='Print debug information')
    parser.add_argument('-x', '--execute', action='store_true', default=True,
                        help="Actually execute. Otherwise, don't do anything.")
    global args  # Help us access this variable from inside the other methods.
    args = parser.parse_args()
    run()

if __name__ == "__main__":
    main(sys.argv[1:])



import unittest

from openformats.utils.json import DumbJson


class DumbJsonTestCase(unittest.TestCase):
    # Dicts
    def test_simple_dict(self):
        self._test_dfs('{"a": "b"}', [("a", 2, "b", 7)])

    def test_simple_dict_with_two_items(self):
        self._test_dfs('{"a": "b", "c": "d"}',
                       [("a", 2, "b", 7), ("c", 12, "d", 17)])

    def test_empty_dict(self):
        self._test_dfs('{  }', [])

    def test_empty_dict_inbetween(self):
        self._test_dfs('["a", {  }, "b"]', [("a", 2), ([], 6), ("b", 13)])

    def test_dict_with_non_string_values(self):
        self._test_dfs('{"string": "Hello World", "True": true, '
                       '"False": false, "None": null, "Integer": 1234, '
                       '"Negative integer": -1234, "Float": 12.34, '
                       '"Negative float": -12.34, "E-notation": 12e34, '
                       '"Negative e-notation": 12e-34}',
                       [("string", 2, "Hello World", 12),
                        ("True", 27, True, 34),
                        ("False", 41, False, 49), ("None", 57, None, 64),
                        ("Integer", 71, 1234, 81),
                        ("Negative integer", 88, -1234, 107),
                        ("Float", 115, 12.34, 123),
                        ("Negative float", 131, -12.34, 148),
                        ("E-notation", 157, 12e34, 170),
                        ("Negative e-notation", 178, 12e-34, 200)])

    # Lists
    def test_simple_list(self):
        self._test_dfs('["a"]', [("a", 2)])

    def test_simple_list_with_two_items(self):
        self._test_dfs('["a", "b"]', [("a", 2), ("b", 7)])

    def test_empty_list(self):
        self._test_dfs('[  ]', [])

    def test_empty_list_inbetween(self):
        self._test_dfs('{"a": "b", "c": [  ], "d": "e"}',
                       [("a", 2, "b", 7), ("c", 12, [], 16),
                        ("d", 23, "e", 28)])

    def test_list_with_non_string_values(self):
        self._test_dfs('["Hello world", true, false, null, 1234, -1234, '
                       '12.34, -12.34, 12e34, 12e-34]',
                       [("Hello world", 2), (True, 16), (False, 22),
                        (None, 29), (1234, 35), (-1234, 41), (12.34, 48),
                        (-12.34, 55), (12e34, 63), (12e-34, 70)])

    # Nested
    def test_dict_within_dict(self):
        self._test_dfs('{"a": {"b": "c"}}', [("a", 2, [("b", 8, "c", 13)], 6)])

    def test_list_within_list(self):
        self._test_dfs('["a", ["b", "c"]]',
                       [("a", 2), ([("b", 8), ("c", 13)], 6)])

    def test_list_within_dict(self):
        self._test_dfs('{"a": ["b", "c"]}',
                       [("a", 2, [("b", 8), ("c", 13)], 6)])

    def test_dict_within_list(self):
        self._test_dfs('["a", {"b": "c"}]',
                       [("a", 2), ([("b", 8, "c", 13)], 6)])

    # Utils
    def _test_dfs(self, content, against):
        dumb_json = DumbJson(content)
        sample = self._dfs(dumb_json)
        self.assertEquals(sample, against)

    def _dfs(self, dumb_json):
        sample = []
        for row in dumb_json:
            if len(row) == 4:  # dict
                key, key_p, value, value_p = row
                if isinstance(value, DumbJson):
                    sample.append((key, key_p,
                                   self._dfs(value), value_p))
                else:
                    sample.append(row)
            elif len(row) == 2:  # list
                item, item_p = row
                if isinstance(item, DumbJson):
                    sample.append((self._dfs(item), item_p))
                else:
                    sample.append(row)
        return sample



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


install_requires = [

]

tests_require = [
    'nose',
    'mock',
    'coverage',
    'nosexcover'
]

setup(
    name="openformats",
    version='0.0.33',
    description="The Transifex Open Formats library",
    author="Transifex",
    author_email="support@transifex.com",
    url="https://github.com/transifex/openformats",
    install_requires=install_requires,
    tests_require=tests_require,
    dependency_links=tests_require,
    test_suite="openformats.tests.run_tests.run_all",
    packages=find_packages(
        where='.',
        exclude=('tests*', 'testbed')
    )
)



from __future__ import absolute_import

import datetime
import inspect
import json
import os
import traceback
from importlib import import_module

from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.views.generic.base import TemplateView, View
from django.views.generic.edit import CreateView

from openformats.handlers import Handler
from openformats.strings import OpenString

from .models import Payload


class HandlerMixin(object):
    @property
    def handlers(self):
        if not hasattr(self, "_handlers"):
            self._handlers = self._get_handlers()
        return self._handlers

    def _get_handlers(self):
        handlers = {}
        format_files = [
            filename.split(".")[0]
            for filename in os.listdir(os.path.join("openformats", "formats"))
            if filename.endswith(".py") and filename != "__init__.py"
        ]
        for filename in format_files:
            module = import_module("openformats.formats.{}".format(filename))
            for name, each in inspect.getmembers(module):
                if (inspect.isclass(each) and issubclass(each, Handler) and
                        each != Handler):
                    if each.name not in handlers:
                        handlers[each.name] = each
        return handlers


class MainView(HandlerMixin, TemplateView):
    http_method_names = ['get']
    template_name = "main/home.html"

    def get(self, request, payload_hash=""):
        self.payload_hash = payload_hash
        return super(MainView, self).get(request, payload_hash=payload_hash)

    def get_context_data(self, **kwargs):
        context = super(MainView, self).get_context_data(**kwargs)
        context['handlers'] = sorted(self.handlers.keys())
        if self.payload_hash:
            payload_row = get_object_or_404(Payload,
                                            payload_hash=self.payload_hash)
            payload_row.last_viewed = datetime.datetime.now()
            payload_row.save()
            context['payload_json'] = payload_row.payload
        return context


class ApiView(HandlerMixin, View):
    def post(self, request):
        payload = json.loads(request.body)
        if payload['action'] == "choose_handler":
            return_value = self._choose_handler(payload)
        elif payload['action'] == "parse":
            return_value = self._parse(payload)
        elif payload['action'] == "compile":
            return_value = self._compile(payload)
        return HttpResponse(json.dumps(return_value),
                            mimetype="application/json")

    def _choose_handler(self, payload):
        handler_name = payload['handler']
        handler_class = self.handlers[handler_name]
        handler_name_lower = handler_name.lower()
        sample_filepath = os.path.join(
            "openformats", "tests", "formats", handler_name_lower, "files",
            "1_en.{}".format(handler_class.extension)
        )
        return_value = {'handler': handler_name}
        try:
            with open(sample_filepath) as f:
                return_value['source'] = f.read()
        except IOError:
            pass
        return return_value

    def _parse(self, payload):
        handler_name = payload['handler']
        handler_class = self.handlers[handler_name]
        source = payload['source']

        handler = handler_class()
        returned_payload = {'action': None, 'compiled': "",
                            'compile_error': ""}
        try:
            template, stringset = handler.parse(source)
        except Exception:
            returned_payload.update({'stringset': [], 'template': "",
                                     'parse_error': traceback.format_exc()})
        else:
            returned_payload.update({
                'stringset': [self._string_to_json(string)
                              for string in stringset],
                'template': template,
                'parse_error': "",
            })

        return returned_payload

    def _string_to_json(self, string):
        return_value = {'id': string.template_replacement,
                        'key': string.key,
                        'strings': string._strings,
                        'pluralized': string.pluralized,
                        'template_replacement': string.template_replacement}
        for key in OpenString.DEFAULTS:
            return_value[key] = getattr(string, key)
        return return_value

    def _compile(self, payload):
        handler_name = payload['handler']
        handler_class = self.handlers[handler_name]
        stringset_json = payload['stringset']
        template = payload['template']

        stringset = []
        for string_json in stringset_json:
            key = string_json.pop('key')
            strings = {int(key): value
                       for key, value in string_json.pop('strings').items()}
            del string_json['pluralized']
            del string_json['template_replacement']
            stringset.append(OpenString(key, strings, **string_json))

        handler = handler_class()
        try:
            compiled = handler.compile(template, stringset)
        except Exception:
            returned_payload = {'action': None, 'compiled': "",
                                'compile_error': traceback.format_exc()}
        else:
            returned_payload = {'action': None, 'compiled': compiled,
                                'compile_error': ""}
        return returned_payload


class SaveView(CreateView):
    http_method_names = ['post']
    model = Payload

    def form_valid(self, form):
        try:
            return super(SaveView, self).form_valid(form)
        except Exception:
            return HttpResponseRedirect(form.instance.get_absolute_url())

    def form_invalid(self, form):
        return HttpResponseRedirect(form.instance.get_absolute_url())
